from dataclasses import dataclass
from typing import List

from osbenchmark.telemetry import Telemetry


@dataclass
class Node:
    """A representation of a node within a host"""

    name: str
    port: int
    pid: int
    root_dir: str
    binary_path: str
    log_path: str
    heap_dump_path: str
    data_paths: List[str]
    telemetry: Telemetry



from osbenchmark.builder.executors.shell_executor import ShellExecutor
from osbenchmark.exceptions import ExecutorError


class ExceptionHandlingShellExecutor(ShellExecutor):
    def __init__(self, executor):
        self.executor = executor

    def execute(self, host, command, **kwargs):
        try:
            return self.executor.execute(host, command, kwargs)
        except Exception as e:
            raise ExecutorError(f"Command \"{command}\" on host \"{host}\" failed to execute", e)



import logging
import os.path

from osbenchmark.builder.downloaders.downloader import Downloader
from osbenchmark.builder.utils.binary_keys import BinaryKeys
from osbenchmark.exceptions import ExecutorError


class OpenSearchDistributionDownloader(Downloader):
    def __init__(self, provision_config_instance, executor, path_manager, distribution_repository_provider):
        super().__init__(executor)
        self.logger = logging.getLogger(__name__)
        self.provision_config_instance = provision_config_instance
        self.path_manager = path_manager
        self.distribution_repository_provider = distribution_repository_provider

    def download(self, host):
        binary_path = self._fetch_binary(host)
        return {BinaryKeys.OPENSEARCH: binary_path}

    def _fetch_binary(self, host):
        download_url = self.distribution_repository_provider.get_download_url(host)
        distribution_path = self._create_distribution_path(host, download_url)
        opensearch_version = self.provision_config_instance.variables["distribution"]["version"]

        is_binary_present = self._is_binary_present(host, distribution_path)
        is_cache_enabled = self.distribution_repository_provider.is_cache_enabled()

        if is_binary_present and is_cache_enabled:
            self.logger.info("Skipping download for version [%s]. Found existing binary at [%s].", opensearch_version,
                             distribution_path)
        else:
            self._download_opensearch(host, distribution_path, download_url, opensearch_version)

        return distribution_path

    def _create_distribution_path(self, host, download_url):
        distribution_root_path = os.path.join(self.provision_config_instance.variables["node"]["root"]["dir"], "distributions")
        self.path_manager.create_path(host, distribution_root_path)

        distribution_binary_name = self.distribution_repository_provider.get_file_name_from_download_url(download_url)
        return os.path.join(distribution_root_path, distribution_binary_name)

    def _is_binary_present(self, host, distribution_path):
        try:
            self.executor.execute(host, f"test -f {distribution_path}")
            return True
        except ExecutorError:
            return False

    def _download_opensearch(self, host, distribution_path, download_url, opensearch_version):
        self.logger.info("Resolved download URL [%s] for version [%s]", download_url, opensearch_version)
        self.logger.info("Starting download of OpenSearch [%s]", opensearch_version)

        try:
            self.executor.execute(host, f"curl -o {distribution_path} {download_url}")
        except ExecutorError as e:
            self.logger.exception("Exception downloading OpenSearch distribution for version [%s] from [%s].",
                                  opensearch_version, download_url)
            raise e

        self.logger.info("Successfully downloaded OpenSearch [%s].", opensearch_version)



from dataclasses import dataclass

from osbenchmark.builder.models.node import Node


@dataclass
class Host:
    """A representation of a host within a cluster"""

    name: str
    address: str
    metadata: dict
    node: Node



from osbenchmark.exceptions import ExecutorError
from osbenchmark.utils import io


class PathManager:
    def __init__(self, executor):
        self.executor = executor

    def create_path(self, host, path, create_locally=True):
        if create_locally:
            io.ensure_dir(path)
        self.executor.execute(host, "mkdir -m 0777 -p " + path)

    def is_path_present(self, host, path):
        try:
            self.executor.execute(host, f"test -e {path}")
            return True
        except ExecutorError:
            return False

    def delete_path(self, host, path):
        path_block_list = ["", "*", "/", None]
        if path in path_block_list:
            return

        self.executor.execute(host, "rm -r " + path)



from abc import ABC, abstractmethod


class Launcher(ABC):
    """
    Launchers are used to start and stop OpenSearch on the nodes in a self-managed cluster.
    """
    def __init__(self, shell_executor):
        self.shell_executor = shell_executor

    @abstractmethod
    def start(self, host, node_configurations):
        """
        Starts the OpenSearch nodes on a given host

        ;param host: A Host object defining the host on which to start the nodes
        ;param node_configurations: A list of NodeConfiguration objects detailing the installation data of the nodes on the host
        ;return nodes: A list of Node objects defining the nodes running on a host
        """
        raise NotImplementedError

    @abstractmethod
    def stop(self, host, nodes):
        """
        Stops the OpenSearch nodes on a given host

        ;param host: A Host object defining the host on which to stop the nodes
        ;param nodes: A list of Node objects defining the nodes running on a host
        ;return nodes: A list of Node objects representing OpenSearch nodes that were successfully stopped on the host
        """
        raise NotImplementedError



# SPDX-License-Identifier: Apache-2.0
#
# The OpenSearch Contributors require contributions made to
# this file be licensed under the Apache-2.0 license or a
# compatible open source license.
# Modifications Copyright OpenSearch Contributors. See
# GitHub history for details.

# Licensed to Elasticsearch B.V. under one or more contributor
# license agreements. See the NOTICE file distributed with
# this work for additional information regarding copyright
# ownership. Elasticsearch B.V. licenses this file to you under
# the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#	http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import logging
import os

from osbenchmark.utils import process
from osbenchmark import exceptions

LOGGER_NAME = "osbenchmark.provisioner.repository_gcs"


def resolve_binary(install_root, binary_name):
    return os.path.join(install_root, "bin", binary_name)


def resolve_keystore_config(install_root):
    return os.path.join(install_root, "config", "opensearch.keystore")


def create_keystore(install_root, keystore_binary, env):
    logger = logging.getLogger(LOGGER_NAME)

    keystore_create_command = "{keystore} --silent create".format(keystore=keystore_binary)

    return_code = process.run_subprocess_with_logging(
        keystore_create_command,
        env=env
    )

    if return_code != 0:
        logger.error("%s has exited with code [%d]", keystore_create_command, return_code)
        raise exceptions.SystemSetupError(
            "Could not initialize a keystore. Please see the log for details.")


def configure_keystore(config_names, variables, **kwargs):
    logger = logging.getLogger(LOGGER_NAME)
    keystore_params = ["gcs_client_name", "gcs_credentials_file"]
    client_name = variables.get(keystore_params[0])
    credentials_file = variables.get(keystore_params[1])

    if not (credentials_file and client_name):
        logger.warning("Skipping keystore configuration for repository-gcs as plugin-params %s were not supplied", keystore_params)
        return False

    keystore_binary_filename = "opensearch-keystore"
    install_root = variables["install_root_path"]
    keystore_binary = resolve_binary(install_root, keystore_binary_filename)
    env = kwargs.get("env")

    if not os.path.isfile(resolve_keystore_config(install_root)):
        create_keystore(install_root, keystore_binary, env)

    keystore_command = "{keystore} --silent add-file gcs.client.{client_name}.credentials_file {credentials_file}".format(
        keystore=keystore_binary,
        client_name=client_name,
        credentials_file=credentials_file)

    return_code = process.run_subprocess_with_logging(
        keystore_command,
        env=env
    )

    if return_code != 0:
        logger.error("%s has exited with code [%d]", keystore_command, return_code)
        raise exceptions.SystemSetupError(
            "Could not add GCS keystore secure setting. Please see the log for details.")

    # Success
    return True


def register(registry):
    registry.register("post_install", configure_keystore)



class PluginDistributionRepositoryProvider:
    def __init__(self, plugin, repository_url_provider):
        self.plugin = plugin
        self.repository_url_provider = repository_url_provider

    def get_download_url(self, host):
        distribution_repository = self.plugin.variables["distribution"]["repository"]

        default_key = f"distribution.{distribution_repository}.remote.repo.url"
        return self.repository_url_provider.render_url_for_key(host, self.plugin.variables, default_key, mandatory=False)



import logging

from osbenchmark.utils import convert


class OpenSearchDistributionRepositoryProvider:
    def __init__(self, provision_config_instance, repository_url_provider):
        self.logger = logging.getLogger(__name__)
        self.provision_config_instance = provision_config_instance
        self.repository_url_provider = repository_url_provider

    def get_download_url(self, host):
        is_runtime_jdk_bundled = self.provision_config_instance.variables["system"]["runtime"]["jdk"]["bundled"]
        distribution_repository = self.provision_config_instance.variables["distribution"]["repository"]

        self.logger.info("runtime_jdk_bundled? [%s]", is_runtime_jdk_bundled)
        if is_runtime_jdk_bundled:
            url_key = f"distribution.jdk.bundled.{distribution_repository}_url"
        else:
            url_key = f"distribution.jdk.unbundled.{distribution_repository}_url"

        self.logger.info("key: [%s]", url_key)
        return self.repository_url_provider.render_url_for_key(host, self.provision_config_instance.variables, url_key)

    def get_file_name_from_download_url(self, download_url):
        return download_url[download_url.rfind("/") + 1:]

    def is_cache_enabled(self):
        distribution_repository = self.provision_config_instance.variables["distribution"]["repository"]
        is_cache_enabled = self.provision_config_instance.variables["distribution"][distribution_repository]["cache"]

        return convert.to_bool(is_cache_enabled)



# SPDX-License-Identifier: Apache-2.0
#
# The OpenSearch Contributors require contributions made to
# this file be licensed under the Apache-2.0 license or a
# compatible open source license.
# Modifications Copyright OpenSearch Contributors. See
# GitHub history for details.
# Licensed to Elasticsearch B.V. under one or more contributor
# license agreements. See the NOTICE file distributed with
# this work for additional information regarding copyright
# ownership. Elasticsearch B.V. licenses this file to you under
# the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#	http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import collections
import numbers
import re
from enum import Enum, unique

from osbenchmark import exceptions


class Index:
    """
    Defines an index in OpenSearch.
    """

    def __init__(self, name, body=None, types=None):
        """

        Creates a new index.

        :param name: The index name. Mandatory.
        :param body: A dict representation of the index body. Optional.
        :param types: A list of types. Should contain at least one type.
        """
        if types is None:
            types = []
        if body is None:
            body = {}
        self.name = name
        self.body = body
        self.types = types

    def matches(self, pattern):
        if pattern is None:
            return True
        elif pattern in ["_all", "*"]:
            return True
        elif self.name == pattern:
            return True
        else:
            return False

    def __str__(self):
        return self.name

    def __repr__(self):
        r = []
        for prop, value in vars(self).items():
            r.append("%s = [%s]" % (prop, repr(value)))
        return ", ".join(r)

    def __hash__(self):
        return hash(self.name)

    def __eq__(self, other):
        return self.name == other.name


class DataStream:
    """
    Defines a data stream in OpenSearch.
    """

    def __init__(self, name):
        """

        Creates a new data stream.

        :param name: The data stream name. Mandatory.
        """
        self.name = name

    def matches(self, pattern):
        if pattern is None:
            return True
        elif pattern in ["_all", "*"]:
            return True
        elif self.name == pattern:
            return True
        else:
            return False

    def __str__(self):
        return self.name

    def __repr__(self):
        r = []
        for prop, value in vars(self).items():
            r.append("%s = [%s]" % (prop, repr(value)))
        return ", ".join(r)

    def __hash__(self):
        return hash(self.name)

    def __eq__(self, other):
        return self.name == other.name


class IndexTemplate:
    """
    Defines an index template in OpenSearch.
    """

    def __init__(self, name, pattern, content, delete_matching_indices=False):
        """

        Creates a new index template.

        :param name: Name of the index template. Mandatory.
        :param pattern: The index pattern to which the index template applies. Mandatory.
        :param content: The content of the corresponding template. Mandatory.
        :param delete_matching_indices: Delete all indices that match the pattern before the benchmark iff True.
        """
        self.name = name
        self.pattern = pattern
        self.content = content
        self.delete_matching_indices = delete_matching_indices

    def __str__(self, *args, **kwargs):
        return self.name

    def __repr__(self):
        r = []
        for prop, value in vars(self).items():
            r.append("%s = [%s]" % (prop, repr(value)))
        return ", ".join(r)

    def __hash__(self):
        return hash(self.name)

    def __eq__(self, other):
        return self.name == other.name


class ComponentTemplate:
    """
    Defines a component template in OpenSearch.
    """

    def __init__(self, name, content):
        """
        Creates a new index template.
        :param name: Name of the index template. Mandatory.
        :param content: The content of the corresponding template. Mandatory.
        """
        self.name = name
        self.content = content

    def __str__(self, *args, **kwargs):
        return self.name

    def __repr__(self):
        r = []
        for prop, value in vars(self).items():
            r.append("%s = [%s]" % (prop, repr(value)))
        return ", ".join(r)

    def __hash__(self):
        return hash(self.name)

    def __eq__(self, other):
        return self.name == other.name


class Documents:
    SOURCE_FORMAT_BULK = "bulk"

    def __init__(self, source_format, document_file=None, document_archive=None, base_url=None,
                 includes_action_and_meta_data=False,
                 number_of_documents=0, compressed_size_in_bytes=0, uncompressed_size_in_bytes=0, target_index=None,
                 target_data_stream=None, target_type=None, meta_data=None):
        """

        :param source_format: The format of these documents. Mandatory.
        :param document_file: The file name of benchmark documents after decompression. Optional (e.g. for percolation we
        just need a mapping but no documents)
        :param document_archive: The file name of the compressed benchmark document name on the remote server. Optional (e.g. for
        percolation we just need a mapping but no documents)
        :param base_url: The URL from which to load data if they are not available locally. Optional.
        :param includes_action_and_meta_data: True, if the source file already includes the action and meta-data line. False, if it only
        contains documents.
        :param number_of_documents: The number of documents
        in the benchmark document. Needed for proper progress results_publishing. Only needed if
         a document_archive is given.
        :param compressed_size_in_bytes: The compressed size in bytes of
        the benchmark document. Needed for verification of the download and
         user results_publishing. Only useful if a document_archive is given (optional but recommended to be set).
        :param uncompressed_size_in_bytes: The size in bytes of the benchmark document after decompressing it.
        Only useful if a document_archive is given (optional but recommended to be set).
        :param target_index: The index to target for bulk operations. May be ``None`` if ``includes_action_and_meta_data`` is ``False``.
        :param target_data_stream: The data stream to target for bulk operations.
        Maybe be ``None`` if ``includes_action_and_meta_data`` is ``False``.
        :param target_type: The document type to target for bulk operations. May be ``None`` if ``includes_action_and_meta_data``
                            is ``False``.
        :param meta_data: A dict containing key-value pairs with additional meta-data describing documents. Optional.
        """

        self.source_format = source_format
        self.document_file = document_file
        self.document_archive = document_archive
        self.base_url = base_url
        self.includes_action_and_meta_data = includes_action_and_meta_data
        self._number_of_documents = number_of_documents
        self._compressed_size_in_bytes = compressed_size_in_bytes
        self._uncompressed_size_in_bytes = uncompressed_size_in_bytes
        self.target_index = target_index
        self.target_data_stream = target_data_stream
        self.target_type = target_type
        self.meta_data = meta_data or {}

    def has_compressed_corpus(self):
        return self.document_archive is not None

    def has_uncompressed_corpus(self):
        return self.document_file is not None

    @property
    def number_of_documents(self):
        return self._number_of_documents

    @number_of_documents.setter
    def number_of_documents(self, value):
        self._number_of_documents = value

    @property
    def uncompressed_size_in_bytes(self):
        return self._uncompressed_size_in_bytes

    @uncompressed_size_in_bytes.setter
    def uncompressed_size_in_bytes(self, value):
        self._uncompressed_size_in_bytes = value

    @property
    def compressed_size_in_bytes(self):
        return self._compressed_size_in_bytes

    @compressed_size_in_bytes.setter
    def compressed_size_in_bytes(self, value):
        self._compressed_size_in_bytes = value

    @property
    def number_of_lines(self):
        if self.includes_action_and_meta_data:
            return self.number_of_documents * 2
        else:
            return self.number_of_documents

    @property
    def is_bulk(self):
        return self.source_format == Documents.SOURCE_FORMAT_BULK

    def __str__(self):
        return "%s documents from %s" % (self.source_format, self.document_file)

    def __repr__(self):
        r = []
        for prop, value in vars(self).items():
            r.append("%s = [%s]" % (prop, repr(value)))
        return ", ".join(r)

    def __hash__(self):
        return hash(self.source_format) ^ hash(self.document_file) ^ hash(self.document_archive) ^ hash(self.base_url) ^ \
               hash(self.includes_action_and_meta_data) ^ hash(self.number_of_documents) ^ hash(self.compressed_size_in_bytes) ^ \
               hash(self.uncompressed_size_in_bytes) ^ hash(self.target_index) ^ hash(self.target_data_stream) ^ hash(self.target_type) ^ \
               hash(frozenset(self.meta_data.items()))

    def __eq__(self, othr):
        return (isinstance(othr, type(self)) and
                (self.source_format, self.document_file, self.document_archive, self.base_url, self.includes_action_and_meta_data,
                 self.number_of_documents, self.compressed_size_in_bytes, self.uncompressed_size_in_bytes,
                 self.target_type, self.target_data_stream, self.target_type, self.meta_data) ==
                (othr.source_format, othr.document_file, othr.document_archive, othr.base_url, othr.includes_action_and_meta_data,
                 othr.number_of_documents, othr.compressed_size_in_bytes, othr.uncompressed_size_in_bytes,
                 othr.target_type, othr.target_data_stream, othr.target_type, othr.meta_data))


class DocumentCorpus:
    def __init__(self, name, documents=None, meta_data=None):
        """

        :param name: The name of this document corpus. Mandatory.
        :param documents: A list of ``Documents`` instances that belong to this corpus.
        :param meta_data: A dict containing key-value pairs with additional meta-data describing this corpus. Optional.
        """
        self.name = name
        self.documents = documents or []
        self.meta_data = meta_data or {}

    def number_of_documents(self, source_format):
        num = 0
        for doc in self.documents:
            if doc.source_format == source_format:
                num += doc.number_of_documents
        return num

    def compressed_size_in_bytes(self, source_format):
        num = 0
        for doc in self.documents:
            if doc.source_format == source_format and doc.compressed_size_in_bytes is not None:
                num += doc.compressed_size_in_bytes
            else:
                return None
        return num

    def uncompressed_size_in_bytes(self, source_format):
        num = 0
        for doc in self.documents:
            if doc.source_format == source_format and doc.uncompressed_size_in_bytes is not None:
                num += doc.uncompressed_size_in_bytes
            else:
                return None
        return num

    def filter(self, source_format=None, target_indices=None, target_data_streams=None):
        filtered = []
        for d in self.documents:
            # skip if source format or target index does not match
            if source_format and d.source_format != source_format:
                continue
            if target_indices and d.target_index not in target_indices:
                continue
            if target_data_streams and d.target_data_stream not in target_data_streams:
                continue

            filtered.append(d)
        return DocumentCorpus(self.name, filtered, meta_data=dict(self.meta_data))

    def union(self, other):
        """
        Creates a new corpus based on the current and the provided other corpus. This is not meant as a generic union
        of two arbitrary corpora but rather to unify the documents referenced by two instances of the same corpus. This
        is useful when two tasks reference different subsets of a corpus and a unified view (e.g. for downloading the
        appropriate document files) is required.

        :param other: The other corpus to unify with this one. Must have the same name and meta-data.
        :return: A document corpus instance with the same and meta-data but with documents from both corpora.
        """
        if self.name != other.name:
            raise exceptions.BenchmarkAssertionError(f"Corpora names differ: [{self.name}] and [{other.name}].")
        if self.meta_data != other.meta_data:
            raise exceptions.BenchmarkAssertionError(f"Corpora meta-data differ: [{self.meta_data}] and [{other.meta_data}].")
        if self is other:
            return self
        else:
            return DocumentCorpus(name=self.name,
                                  documents=list(set(self.documents).union(other.documents)),
                                  meta_data=dict(self.meta_data))

    def __str__(self):
        return self.name

    def __repr__(self):
        r = []
        for prop, value in vars(self).items():
            r.append("%s = [%s]" % (prop, repr(value)))
        return ", ".join(r)

    def __hash__(self):
        return hash(self.name) ^ hash(self.documents) ^ hash(frozenset(self.meta_data.items()))

    def __eq__(self, othr):
        return (isinstance(othr, type(self)) and
                (self.name, self.documents, self.meta_data) ==
                (othr.name, othr.documents, othr.meta_data))


class Workload:
    """
    A workload defines the data set that is used. It corresponds loosely to a use case (e.g. logging, event processing, analytics, ...)
    """

    def __init__(self, name, description=None, meta_data=None, test_procedures=None, indices=None, data_streams=None,
                 templates=None, composable_templates=None, component_templates=None, corpora=None, has_plugins=False):
        """

        Creates a new workload.

        :param name: A short, descriptive name for this workload. As per convention, this name should be in lower-case without spaces.
        :param description: A description for this workload (should be less than 80 characters).
        :param meta_data: An optional dict of meta-data elements to attach to each metrics record. Default: {}.
        :param test_procedures: A list of one or more test_procedures to use.
        Precondition: If the list is non-empty it contains exactly one element
        with its ``default`` property set to ``True``.
        :param indices: A list of indices for this workload. May be None.
        :param data_streams: A list of data streams for this workload. May be None.
        :param templates: A list of index templates for this workload. May be None.
        :param corpora: A list of document corpus definitions for this workload. May be None.
        :param has_plugins: True iff the workload also defines plugins (e.g. custom runners or parameter sources).
        """
        self.name = name
        self.meta_data = meta_data if meta_data else {}
        self.description = description if description is not None else ""
        self.test_procedures = test_procedures if test_procedures else []
        self.indices = indices if indices else []
        self.data_streams = data_streams if data_streams else []
        self.corpora = corpora if corpora else []
        self.templates = templates if templates else []
        self.composable_templates = composable_templates if composable_templates else []
        self.component_templates = component_templates if component_templates else []
        self.has_plugins = has_plugins

    @property
    def default_test_procedure(self):
        for test_procedure in self.test_procedures:
            if test_procedure.default:
                return test_procedure
        # This should only happen if we don't have any test_procedures
        return None

    @property
    def selected_test_procedure(self):
        for test_procedure in self.test_procedures:
            if test_procedure.selected:
                return test_procedure
        return None

    @property
    def selected_test_procedure_or_default(self):
        selected = self.selected_test_procedure
        return selected if selected else self.default_test_procedure

    def find_test_procedure_or_default(self, name):
        """
        :param name: The name of the test_procedure to find.
        :return: The test_procedure with the given name. The default test_procedure, if the name is "" or ``None``.
        """
        if name in [None, ""]:
            return self.default_test_procedure
        else:
            return self.find_test_procedure(name)

    def find_test_procedure(self, name):
        for test_procedure in self.test_procedures:
            if test_procedure.name == name:
                return test_procedure
        raise exceptions.InvalidName("Unknown test_procedure [%s] for workload [%s]" % (name, self.name))

    @property
    def number_of_documents(self):
        num_docs = 0
        for corpus in self.corpora:
            # TODO #341: Improve API to let users define what they want (everything, just specific types, ...)
            num_docs += corpus.number_of_documents(Documents.SOURCE_FORMAT_BULK)
        return num_docs

    @property
    def compressed_size_in_bytes(self):
        size = 0
        for corpus in self.corpora:
            # TODO #341: Improve API to let users define what they want (everything, just specific types, ...)
            curr_size = corpus.compressed_size_in_bytes(Documents.SOURCE_FORMAT_BULK)
            if curr_size is not None:
                size += curr_size
            else:
                return None
        return size

    @property
    def uncompressed_size_in_bytes(self):
        size = 0
        for corpus in self.corpora:
            # TODO #341: Improve API to let users define what they want (everything, just specific types, ...)
            curr_size = corpus.uncompressed_size_in_bytes(Documents.SOURCE_FORMAT_BULK)
            if curr_size is not None:
                size += curr_size
            else:
                return None
        return size

    def __str__(self):
        return self.name

    def __repr__(self):
        r = []
        for prop, value in vars(self).items():
            r.append("%s = [%s]" % (prop, repr(value)))
        return ", ".join(r)

    def __hash__(self):
        return hash(self.name) ^ hash(self.meta_data) ^ hash(self.description) ^ hash(self.test_procedures) ^ \
               hash(self.indices) ^ hash(self.templates) ^ hash(self.composable_templates) ^ hash(self.component_templates) \
               ^ hash(self.corpora)

    def __eq__(self, othr):
        return (isinstance(othr, type(self)) and
                (self.name, self.meta_data, self.description, self.test_procedures, self.indices, self.data_streams,
                 self.templates, self.composable_templates, self.component_templates, self.corpora) ==
                (othr.name, othr.meta_data, othr.description, othr.test_procedures, othr.indices, othr.data_streams,
                 othr.templates, othr.composable_templates, othr.component_templates, othr.corpora))


class TestProcedure:
    """
    A test procedure defines the concrete operations that will be done.
    """
    #Pytest throws a collection warning if the following line is removed
    __test__ = False
    def __init__(self,
                 name,
                 description=None,
                 user_info=None,
                 default=False,
                 selected=False,
                 auto_generated=False,
                 parameters=None,
                 meta_data=None,
                 schedule=None):
        self.name = name
        self.parameters = parameters if parameters else {}
        self.meta_data = meta_data if meta_data else {}
        self.description = description
        self.user_info = user_info
        self.default = default
        self.selected = selected
        self.auto_generated = auto_generated
        self.schedule = schedule if schedule else []

    def prepend_tasks(self, tasks):
        self.schedule = tasks + self.schedule

    def remove_task(self, task):
        self.schedule.remove(task)

    def __str__(self):
        return self.name

    def __repr__(self):
        r = []
        for prop, value in vars(self).items():
            r.append("%s = [%s]" % (prop, repr(value)))
        return ", ".join(r)

    def __hash__(self):
        return hash(self.name) ^ hash(self.description) ^ hash(self.default) ^ \
               hash(self.selected) ^ hash(self.auto_generated) ^ hash(self.parameters) ^ hash(self.meta_data) ^ \
               hash(self.schedule)

    def __eq__(self, othr):
        return (isinstance(othr, type(self)) and
                (self.name, self.description, self.default, self.selected, self.auto_generated,
                 self.parameters, self.meta_data, self.schedule) ==
                (othr.name, othr.description, othr.default, othr.selected, othr.auto_generated,
                 othr.parameters, othr.meta_data, othr.schedule))


@unique
class OperationType(Enum):
    # for the time being we are not considering this action as administrative
    IndexStats = 1
    NodeStats = 2
    Search = 3
    Bulk = 4
    RawRequest = 5
    WaitForRecovery = 6
    WaitForSnapshotCreate = 7
    Composite = 8
    SubmitAsyncSearch = 9
    GetAsyncSearch = 10
    DeleteAsyncSearch = 11
    PaginatedSearch = 12
    ScrollSearch = 13
    CreatePointInTime = 14
    DeletePointInTime = 15
    ListAllPointInTime = 16

    # administrative actions
    ForceMerge = 1001
    ClusterHealth = 1002
    PutPipeline = 1003
    Refresh = 1004
    CreateIndex = 1005
    DeleteIndex = 1006
    CreateIndexTemplate = 1007
    DeleteIndexTemplate = 1008
    ShrinkIndex = 1009
    Sleep = 1018
    DeleteSnapshotRepository = 1019
    CreateSnapshotRepository = 1020
    CreateSnapshot = 1021
    RestoreSnapshot = 1022
    PutSettings = 1023
    CreateTransform = 1024
    StartTransform = 1025
    WaitForTransform = 1026
    DeleteTransform = 1027
    CreateDataStream = 1028
    DeleteDataStream = 1029
    CreateComposableTemplate = 1030
    DeleteComposableTemplate = 1031
    CreateComponentTemplate = 1032
    DeleteComponentTemplate = 1033

    @property
    def admin_op(self):
        # pylint: disable=comparison-with-callable
        return self.value > 1000

    def to_hyphenated_string(self):
        """
        Turns enum constants into hyphenated names, e.g. ``WaitForTransform`` becomes ``wait-for-transform``.
        """
        # Pylint complains that self.name is not iterable
        # pylint: disable=not-an-iterable
        return "".join(["-" + c.lower() if c.isupper() else c for c in self.name]).lstrip("-")

    # pylint: disable=too-many-return-statements
    @classmethod
    def from_hyphenated_string(cls, v):
        if v == "force-merge":
            return OperationType.ForceMerge
        elif v == "index-stats":
            return OperationType.IndexStats
        elif v == "node-stats":
            return OperationType.NodeStats
        elif v == "search":
            return OperationType.Search
        elif v == "scroll-search":
            return OperationType.ScrollSearch
        elif v == "paginated-search":
            return OperationType.PaginatedSearch
        elif v == "cluster-health":
            return OperationType.ClusterHealth
        elif v == "bulk":
            return OperationType.Bulk
        elif v == "raw-request":
            return OperationType.RawRequest
        elif v == "put-pipeline":
            return OperationType.PutPipeline
        elif v == "refresh":
            return OperationType.Refresh
        elif v == "create-index":
            return OperationType.CreateIndex
        elif v == "delete-index":
            return OperationType.DeleteIndex
        elif v == "create-index-template":
            return OperationType.CreateIndexTemplate
        elif v == "delete-index-template":
            return OperationType.DeleteIndexTemplate
        elif v == "create-composable-template":
            return OperationType.CreateComposableTemplate
        elif v == "delete-composable-template":
            return OperationType.DeleteComposableTemplate
        elif v == "create-component-template":
            return OperationType.CreateComponentTemplate
        elif v == "delete-component-template":
            return OperationType.DeleteComponentTemplate
        elif v == "shrink-index":
            return OperationType.ShrinkIndex
        elif v == "sleep":
            return OperationType.Sleep
        elif v == "delete-snapshot-repository":
            return OperationType.DeleteSnapshotRepository
        elif v == "create-snapshot-repository":
            return OperationType.CreateSnapshotRepository
        elif v == "create-snapshot":
            return OperationType.CreateSnapshot
        elif v == "wait-for-snapshot-create":
            return OperationType.WaitForSnapshotCreate
        elif v == "restore-snapshot":
            return OperationType.RestoreSnapshot
        elif v == "wait-for-recovery":
            return OperationType.WaitForRecovery
        elif v == "put-settings":
            return OperationType.PutSettings
        elif v == "create-transform":
            return OperationType.CreateTransform
        elif v == "start-transform":
            return OperationType.StartTransform
        elif v == "wait-for-transform":
            return OperationType.WaitForTransform
        elif v == "delete-transform":
            return OperationType.DeleteTransform
        elif v == "create-data-stream":
            return OperationType.CreateDataStream
        elif v == "delete-data-stream":
            return OperationType.DeleteDataStream
        elif v == "composite":
            return OperationType.Composite
        elif v == "submit-async-search":
            return OperationType.SubmitAsyncSearch
        elif v == "get-async-search":
            return OperationType.GetAsyncSearch
        elif v == "delete-async-search":
            return OperationType.DeleteAsyncSearch
        elif v == "create-point-in-time":
            return OperationType.CreatePointInTime
        elif v == "delete-point-in-time":
            return OperationType.DeletePointInTime
        elif v == "list-all-point-in-time":
            return OperationType.ListAllPointInTime
        else:
            raise KeyError(f"No enum value for [{v}]")


class TaskNameFilter:
    def __init__(self, name):
        self.name = name

    def matches(self, task):
        return self.name == task.name

    def __hash__(self):
        return hash(self.name)

    def __eq__(self, other):
        return isinstance(other, type(self)) and self.name == other.name

    def __str__(self, *args, **kwargs):
        return f"filter for task name [{self.name}]"


class TaskOpTypeFilter:
    def __init__(self, op_type_name):
        self.op_type = op_type_name

    def matches(self, task):
        return self.op_type == task.operation.type

    def __hash__(self):
        return hash(self.op_type)

    def __eq__(self, other):
        return isinstance(other, type(self)) and self.op_type == other.op_type

    def __str__(self, *args, **kwargs):
        return f"filter for operation type [{self.op_type}]"


class TaskTagFilter:
    def __init__(self, tag_name):
        self.tag_name = tag_name

    def matches(self, task):
        return self.tag_name in task.tags

    def __hash__(self):
        return hash(self.tag_name)

    def __eq__(self, other):
        return isinstance(other, type(self)) and self.tag_name == other.tag_name

    def __str__(self, *args, **kwargs):
        return f"filter for tasks tagged [{self.tag_name}]"


class Singleton(type):
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]


# Schedule elements
class Parallel:
    def __init__(self, tasks, clients=None):
        self.tasks = tasks
        self._clients = clients
        self.nested = True

    @property
    def clients(self):
        if self._clients is not None:
            return self._clients
        else:
            num_clients = 0
            for task in self.tasks:
                num_clients += task.clients
            return num_clients

    def matches(self, task_filter):
        # a parallel element matches if any of its elements match
        for task in self.tasks:
            if task.matches(task_filter):
                return True
        return False

    def remove_task(self, task):
        self.tasks.remove(task)

    def __iter__(self):
        return iter(self.tasks)

    def __str__(self, *args, **kwargs):
        return "%d parallel tasks" % len(self.tasks)

    def __repr__(self):
        r = []
        for prop, value in vars(self).items():
            r.append("%s = [%s]" % (prop, repr(value)))
        return ", ".join(r)

    def __hash__(self):
        return hash(self.tasks)

    def __eq__(self, other):
        return isinstance(other, type(self)) and self.tasks == other.tasks


Throughput = collections.namedtuple("Throughput", ["value", "unit"])


class Task:
    THROUGHPUT_PATTERN = re.compile(r"(?P<value>(\d*\.)?\d+)\s(?P<unit>\w+/s)")
    IGNORE_RESPONSE_ERROR_LEVEL_WHITELIST = ["non-fatal"]

    def __init__(self, name, operation, tags=None, meta_data=None, warmup_iterations=None, iterations=None,
                 warmup_time_period=None, time_period=None, clients=1, completes_parent=False, schedule=None, params=None):
        self.name = name
        self.operation = operation
        if isinstance(tags, str):
            self.tags = [tags]
        elif tags:
            self.tags = tags
        else:
            self.tags = []
        self.meta_data = meta_data if meta_data else {}
        self.warmup_iterations = warmup_iterations
        self.iterations = iterations
        self.warmup_time_period = warmup_time_period
        self.time_period = time_period
        self.clients = clients
        self.completes_parent = completes_parent
        self.schedule = schedule
        self.params = params if params else {}
        self.nested = False

    def matches(self, task_filter):
        return task_filter.matches(self)

    @property
    def target_throughput(self):
        def numeric(v):
            # While booleans can be converted to a number (False -> 0, True -> 1), we don't want to allow that here
            return isinstance(v, numbers.Number) and not isinstance(v, bool)

        target_throughput = self.params.get("target-throughput")
        target_interval = self.params.get("target-interval")

        if target_interval is not None and target_throughput is not None:
            raise exceptions.InvalidSyntax(f"Task [{self}] specifies target-interval [{target_interval}] and "
                                           f"target-throughput [{target_throughput}] but only one of them is allowed.")

        value = None
        unit = "ops/s"

        if target_interval:
            if not numeric(target_interval):
                raise exceptions.InvalidSyntax(f"Target interval [{target_interval}] for task [{self}] must be numeric.")
            value = 1 / float(target_interval)
        elif target_throughput:
            if isinstance(target_throughput, str):
                matches = re.match(Task.THROUGHPUT_PATTERN, target_throughput)
                if matches:
                    value = float(matches.group("value"))
                    unit = matches.group("unit")
                else:
                    raise exceptions.InvalidSyntax(f"Task [{self}] specifies invalid target throughput [{target_throughput}].")
            elif numeric(target_throughput):
                value = float(target_throughput)
            else:
                raise exceptions.InvalidSyntax(f"Target throughput [{target_throughput}] for task [{self}] "
                                               f"must be string or numeric.")

        if value:
            return Throughput(value, unit)
        else:
            return None

    @property
    def ignore_response_error_level(self):
        ignore_response_error_level = self.params.get("ignore-response-error-level")

        if ignore_response_error_level and \
                ignore_response_error_level not in Task.IGNORE_RESPONSE_ERROR_LEVEL_WHITELIST:
            raise exceptions.InvalidSyntax(
                f"Task [{self}] specifies ignore-response-error-level to [{ignore_response_error_level}] but "
                f"the only allowed values are [{','.join(Task.IGNORE_RESPONSE_ERROR_LEVEL_WHITELIST)}].")

        return ignore_response_error_level

    def error_behavior(self, default_error_behavior):
        """
        Returns the desired behavior when encountering errors during task execution.

        :param default_error_behavior: (str) the default error behavior for the benchmark
        :return: (str) prescribing error handling when a non-fatal error occurs:
            "abort": will fail when any error gets encountered
            "continue": will continue for non fatal errors
        """

        behavior = "continue"
        if default_error_behavior == "abort":
            if self.ignore_response_error_level != "non-fatal":
                behavior = "abort"

        return behavior

    def __hash__(self):
        # Note that we do not include `params` in __hash__ and __eq__ (the other attributes suffice to uniquely define a task)
        return hash(self.name) ^ hash(self.operation) ^ hash(self.warmup_iterations) ^ hash(self.iterations) ^ \
               hash(self.warmup_time_period) ^ hash(self.time_period) ^ hash(self.clients) ^ hash(self.schedule) ^ \
               hash(self.completes_parent)

    def __eq__(self, other):
        # Note that we do not include `params` in __hash__ and __eq__ (the other attributes suffice to uniquely define a task)
        return isinstance(other, type(self)) and (self.name, self.operation, self.warmup_iterations, self.iterations,
                                                  self.warmup_time_period, self.time_period, self.clients, self.schedule,
                                                  self.completes_parent) == (other.name, other.operation, other.warmup_iterations,
                                                                             other.iterations, other.warmup_time_period, other.time_period,
                                                                             other.clients, other.schedule, other.completes_parent)

    def __iter__(self):
        return iter([self])

    def __str__(self, *args, **kwargs):
        return self.name

    def __repr__(self):
        r = []
        for prop, value in vars(self).items():
            r.append("%s = [%s]" % (prop, repr(value)))
        return ", ".join(r)


class Operation:
    def __init__(self, name, operation_type, meta_data=None, params=None, param_source=None):
        if params is None:
            params = {}
        self.name = name
        self.meta_data = meta_data if meta_data else {}
        self.type = operation_type
        self.params = params
        self.param_source = param_source

    @property
    def include_in_results_publishing(self):
        return self.params.get("include-in-results_publishing", True)

    def __hash__(self):
        return hash(self.name)

    def __eq__(self, other):
        return isinstance(other, type(self)) and self.name == other.name

    def __str__(self, *args, **kwargs):
        return self.name

    def __repr__(self):
        r = []
        for prop, value in vars(self).items():
            r.append("%s = [%s]" % (prop, repr(value)))
        return ", ".join(r)



from osbenchmark.builder.launchers.launcher import Launcher
from osbenchmark.exceptions import LaunchError


class ExceptionHandlingLauncher(Launcher):
    def __init__(self, launcher, shell_executor=None):
        super().__init__(shell_executor)
        self.launcher = launcher

    def start(self, host, node_configurations):
        try:
            return self.launcher.start(host, node_configurations)
        except Exception as e:
            raise LaunchError(f"Starting node(s) on host \"{host}\" failed", e)

    def stop(self, host, nodes):
        try:
            return self.launcher.stop(host, nodes)
        except Exception as e:
            raise LaunchError(f"Stopping node(s) on host \"{host}\" failed", e)



from enum import Enum


class ConfigInstanceTypes(Enum):
    """
    A ConfigInstanceType is a representation of a configuration category.

    :param config_type: The type of configuration. This corresponds to the subdirectory name where the configurations are stored
    :param supported_config_format_versions: Multiple formats can be defined for the same configuration type. These non-equal
                                             formats are correlated with a version number. ``supported_config_format_versions``
                                             defines the supported version numbers, which is used for listing available configs
    :param default_config_format_version: The default config format version to use when parsing a configuration. This version
                                          will be used if no corresponding CLI format version is specified by the user
    """

    def __init__(self, config_type, supported_config_format_versions, default_config_format_version):
        self.config_type = config_type
        self.supported_config_format_versions = supported_config_format_versions
        self.default_config_format_version = default_config_format_version

    PLUGIN = "plugins", [1], 1



# SPDX-License-Identifier: Apache-2.0
#
# The OpenSearch Contributors require contributions made to
# this file be licensed under the Apache-2.0 license or a
# compatible open source license.
# Modifications Copyright OpenSearch Contributors. See
# GitHub history for details.
# Licensed to Elasticsearch B.V. under one or more contributor
# license agreements. See the NOTICE file distributed with
# this work for additional information regarding copyright
# ownership. Elasticsearch B.V. licenses this file to you under
# the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#	http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import glob
import json
import logging
import uuid

from osbenchmark import workload, config, exceptions
from osbenchmark.utils import io, console

color_scheme_rgba = [
    # #00BFB3
    "rgba(0,191,179,1)",
    # #00A9E0
    "rgba(0,169,224,1)",
    # #F04E98
    "rgba(240,78,152,1)",
    # #FFCD00
    "rgba(255,205,0,1)",
    # #0076A8
    "rgba(0,118,168,1)",
    # #93C90E
    "rgba(147,201,14,1)",
    # #646464
    "rgba(100,100,100,1)",
]


def index_label(test_execution_config):
    if test_execution_config.label:
        return test_execution_config.label

    label = "%s-%s" % (test_execution_config.test_procedure, test_execution_config.provision_config_instance)
    if test_execution_config.plugins:
        label += "-%s" % test_execution_config.plugins.replace(":", "-").replace(",", "+")
    if test_execution_config.node_count > 1:
        label += " (%d nodes)" % test_execution_config.node_count
    return label


class BarCharts:
    UI_STATE_JSON = json.dumps(
        {
            "vis": {
                "colors": dict(
                    zip(["bare-oss", "bare-basic", "bare-trial-security", "docker-basic", "ear-basic"], color_scheme_rgba))
            }
        })

    @staticmethod
    # flavor's unused but we need the same signature used by the corresponding method in TimeSeriesCharts
    def format_title(environment, workload_name, flavor=None, os_license=None, suffix=None):
        title = f"{environment}-{workload_name}"

        if suffix:
            title += f"-{suffix}"

        return title

    @staticmethod
    def filter_string(environment, test_ex_config):
        if test_ex_config.name:
            return f"environment:\"{environment}\" AND active:true AND user-tags.name:\"{test_ex_config.name}\""
        else:
            return f"environment:\"{environment}\" AND active:true AND workload:\"{test_ex_config.workload}\""\
                   f" AND test_procedure:\"{test_ex_config.test_procedure}\""\
                   f" AND provision_config_instance:\"{test_ex_config.provision_config_instance}\""\
                   f" AND node-count:{test_ex_config.node_count}"

    @staticmethod
    def gc(title, environment, test_execution_config):
        vis_state = {
            "title": title,
            "type": "histogram",
            "params": {
                "addLegend": True,
                "addTimeMarker": False,
                "addTooltip": True,
                "categoryAxes": [
                    {
                        "id": "CategoryAxis-1",
                        "labels": {
                            "show": True,
                            "truncate": 100
                        },
                        "position": "bottom",
                        "scale": {
                            "type": "linear"
                        },
                        "show": True,
                        "style": {},
                        "title": {
                            "text": "filters"
                        },
                        "type": "category"
                    }
                ],
                "defaultYExtents": False,
                "drawLinesBetweenPoints": True,
                "grid": {
                    "categoryLines": False,
                    "style": {
                        "color": "#eee"
                    }
                },
                "interpolate": "linear",
                "legendPosition": "right",
                "radiusRatio": 9,
                "scale": "linear",
                "seriesParams": [
                    {
                        "data": {
                            "id": "1",
                            "label": "Total GC Duration [ms]"
                        },
                        "drawLinesBetweenPoints": True,
                        "mode": "normal",
                        "show": "True",
                        "showCircles": True,
                        "type": "histogram",
                        "valueAxis": "ValueAxis-1"
                    }
                ],
                "setYExtents": False,
                "showCircles": True,
                "times": [],
                "valueAxes": [
                    {
                        "id": "ValueAxis-1",
                        "labels": {
                            "filter": False,
                            "rotate": 0,
                            "show": True,
                            "truncate": 100
                        },
                        "name": "LeftAxis-1",
                        "position": "left",
                        "scale": {
                            "mode": "normal",
                            "type": "linear"
                        },
                        "show": True,
                        "style": {},
                        "title": {
                            "text": "Total GC Duration [ms]"
                        },
                        "type": "value"
                    }
                ]
            },
            "aggs": [
                {
                    "id": "1",
                    "enabled": True,
                    "type": "median",
                    "schema": "metric",
                    "params": {
                        "field": "value.single",
                        "percents": [
                            50
                        ],
                        "customLabel": "Total GC Duration [ms]"
                    }
                },
                {
                    "id": "2",
                    "enabled": True,
                    "type": "filters",
                    "schema": "segment",
                    "params": {
                        "filters": [
                            {
                                "input": {
                                    "query": {
                                        "query_string": {
                                            "query": "name:young_gc_time",
                                            "analyze_wildcard": True
                                        }
                                    }
                                },
                                "label": "Young GC"
                            },
                            {
                                "input": {
                                    "query": {
                                        "query_string": {
                                            "query": "name:old_gc_time",
                                            "analyze_wildcard": True
                                        }
                                    }
                                },
                                "label": "Old GC"
                            }
                        ]
                    }
                },
                {
                    "id": "3",
                    "enabled": True,
                    "type": "terms",
                    "schema": "split",
                    "params": {
                        "field": "distribution-version",
                        "size": 10,
                        "order": "asc",
                        "orderBy": "_term",
                        "row": False
                    }
                },
                {
                    "id": "4",
                    "enabled": True,
                    "type": "terms",
                    "schema": "group",
                    "params": {
                        "field": "user-tags.setup",
                        "size": 5,
                        "order": "desc",
                        "orderBy": "_term"
                    }
                }
            ],
            "listeners": {}
        }

        search_source = {
            "index": "benchmark-results-*",
            "query": {
                "query_string": {
                    "query": BarCharts.filter_string(environment, test_execution_config),
                    "analyze_wildcard": True
                }
            },
            "filter": []
        }

        return {
            "id": str(uuid.uuid4()),
            "type": "visualization",
            "attributes": {
                "title": title,
                "visState": json.dumps(vis_state),
                "uiStateJSON": BarCharts.UI_STATE_JSON,
                "description": "gc",
                "version": 1,
                "kibanaSavedObjectMeta": {
                    "searchSourceJSON": json.dumps(search_source)
                }
            }
        }

    @staticmethod
    def io(title, environment, test_execution_config):
        vis_state = {
            "title": title,
            "type": "histogram",
            "params": {
                "addLegend": True,
                "addTimeMarker": False,
                "addTooltip": True,
                "categoryAxes": [
                    {
                        "id": "CategoryAxis-1",
                        "labels": {
                            "show": True,
                            "truncate": 100
                        },
                        "position": "bottom",
                        "scale": {
                            "type": "linear"
                        },
                        "show": True,
                        "style": {},
                        "title": {
                            "text": "filters"
                        },
                        "type": "category"
                    }
                ],
                "defaultYExtents": False,
                "drawLinesBetweenPoints": True,
                "grid": {
                    "categoryLines": False,
                    "style": {
                        "color": "#eee"
                    }
                },
                "interpolate": "linear",
                "legendPosition": "right",
                "radiusRatio": 9,
                "scale": "linear",
                "seriesParams": [
                    {
                        "data": {
                            "id": "1",
                            "label": "[Bytes]"
                        },
                        "drawLinesBetweenPoints": True,
                        "mode": "normal",
                        "show": "True",
                        "showCircles": True,
                        "type": "histogram",
                        "valueAxis": "ValueAxis-1"
                    }
                ],
                "setYExtents": False,
                "showCircles": True,
                "times": [],
                "valueAxes": [
                    {
                        "id": "ValueAxis-1",
                        "labels": {
                            "filter": False,
                            "rotate": 0,
                            "show": True,
                            "truncate": 100
                        },
                        "name": "LeftAxis-1",
                        "position": "left",
                        "scale": {
                            "mode": "normal",
                            "type": "linear"
                        },
                        "show": True,
                        "style": {},
                        "title": {
                            "text": "[Bytes]"
                        },
                        "type": "value"
                    }
                ]
            },
            "aggs": [
                {
                    "id": "1",
                    "enabled": True,
                    "type": "sum",
                    "schema": "metric",
                    "params": {
                        "field": "value.single",
                        "customLabel": "[Bytes]"
                    }
                },
                {
                    "id": "2",
                    "enabled": True,
                    "type": "filters",
                    "schema": "segment",
                    "params": {
                        "filters": [
                            {
                                "input": {
                                    "query": {
                                        "query_string": {
                                            "analyze_wildcard": True,
                                            "query": "name:index_size"
                                        }
                                    }
                                },
                                "label": "Index size"
                            },
                            {
                                "input": {
                                    "query": {
                                        "query_string": {
                                            "analyze_wildcard": True,
                                            "query": "name:bytes_written"
                                        }
                                    }
                                },
                                "label": "Bytes written"
                            }
                        ]
                    }
                },
                {
                    "id": "3",
                    "enabled": True,
                    "type": "terms",
                    "schema": "split",
                    "params": {
                        "field": "distribution-version",
                        "size": 10,
                        "order": "asc",
                        "orderBy": "_term",
                        "row": False
                    }
                },
                {
                    "id": "4",
                    "enabled": True,
                    "type": "terms",
                    "schema": "group",
                    "params": {
                        "field": "user-tags.setup",
                        "size": 5,
                        "order": "desc",
                        "orderBy": "_term"
                    }
                }
            ],
            "listeners": {}
        }

        search_source = {
            "index": "benchmark-results-*",
            "query": {
                "query_string": {
                    "query": BarCharts.filter_string(environment, test_execution_config),
                    "analyze_wildcard": True
                }
            },
            "filter": []
        }

        return {
            "id": str(uuid.uuid4()),
            "type": "visualization",
            "attributes": {
                "title": title,
                "visState": json.dumps(vis_state),
                "uiStateJSON": BarCharts.UI_STATE_JSON,
                "description": "io",
                "version": 1,
                "kibanaSavedObjectMeta": {
                    "searchSourceJSON": json.dumps(search_source)
                }
            }
        }

    @staticmethod
    def segment_memory(title, environment, test_execution_config):
        # don't generate segment memory charts for releases
        return None

    @staticmethod
    def query(environment, test_execution_config, q):
        metric = "service_time"
        title = BarCharts.format_title(
            environment,
            test_execution_config.workload,
            suffix="%s-%s-p99-%s" % (test_execution_config.label,
            q,
            metric))
        label = "Query Service Time [ms]"

        vis_state = {
            "title": title,
            "type": "histogram",
            "params": {
                "addLegend": True,
                "addTimeMarker": False,
                "addTooltip": True,
                "categoryAxes": [
                    {
                        "id": "CategoryAxis-1",
                        "labels": {
                            "show": True,
                            "truncate": 100
                        },
                        "position": "bottom",
                        "scale": {
                            "type": "linear"
                        },
                        "show": True,
                        "style": {},
                        "title": {
                            "text": "distribution-version: Ascending"
                        },
                        "type": "category"
                    }
                ],
                "defaultYExtents": False,
                "drawLinesBetweenPoints": True,
                "grid": {
                    "categoryLines": False,
                    "style": {
                        "color": "#eee"
                    }
                },
                "interpolate": "linear",
                "legendPosition": "right",
                "radiusRatio": 9,
                "scale": "linear",
                "seriesParams": [
                    {
                        "data": {
                            "id": "1",
                            "label": label
                        },
                        "drawLinesBetweenPoints": True,
                        "mode": "normal",
                        "show": "True",
                        "showCircles": True,
                        "type": "histogram",
                        "valueAxis": "ValueAxis-1"
                    }
                ],
                "setYExtents": False,
                "showCircles": True,
                "times": [],
                "valueAxes": [
                    {
                        "id": "ValueAxis-1",
                        "labels": {
                            "filter": False,
                            "rotate": 0,
                            "show": True,
                            "truncate": 100
                        },
                        "name": "LeftAxis-1",
                        "position": "left",
                        "scale": {
                            "mode": "normal",
                            "type": "linear"
                        },
                        "show": True,
                        "style": {},
                        "title": {
                            "text": label
                        },
                        "type": "value"
                    }
                ]
            },
            "aggs": [
                {
                    "id": "1",
                    "enabled": True,
                    "type": "median",
                    "schema": "metric",
                    "params": {
                        "field": "value.99_0",
                        "percents": [
                            50
                        ],
                        "customLabel": label
                    }
                },
                {
                    "id": "2",
                    "enabled": True,
                    "type": "terms",
                    "schema": "segment",
                    "params": {
                        "field": "distribution-version",
                        "size": 10,
                        "order": "asc",
                        "orderBy": "_term"
                    }
                },
                {
                    "id": "3",
                    "enabled": True,
                    "type": "terms",
                    "schema": "group",
                    "params": {
                        "field": "user-tags.setup",
                        "size": 10,
                        "order": "desc",
                        "orderBy": "_term"
                    }
                }
            ],
            "listeners": {}
        }

        search_source = {
            "index": "benchmark-results-*",
            "query": {
                "query_string": {
                    "query": "name:\"%s\" AND task:\"%s\" AND %s" % (
                        metric,
                        q,
                        BarCharts.filter_string(
                            environment,
                            test_execution_config)),
                    "analyze_wildcard": True
                }
            },
            "filter": []
        }

        return {
            "id": str(uuid.uuid4()),
            "type": "visualization",
            "attributes": {
                "title": title,
                "visState": json.dumps(vis_state),
                "uiStateJSON": BarCharts.UI_STATE_JSON,
                "description": "query",
                "version": 1,
                "kibanaSavedObjectMeta": {
                    "searchSourceJSON": json.dumps(search_source)
                }
            }
        }

    @staticmethod
    def index(environment, test_execution_configs, title):
        filters = []
        for test_execution_config in test_execution_configs:
            label = index_label(test_execution_config)
            # the assumption is that we only have one bulk task
            for bulk_task in test_execution_config.bulk_tasks:
                filters.append({
                    "input": {
                        "query": {
                            "query_string": {
                                "analyze_wildcard": True,
                                "query": "task:\"%s\" AND %s" % (bulk_task, BarCharts.filter_string(environment, test_execution_config))
                            }
                        }
                    },
                    "label": label
                })

        vis_state = {
            "aggs": [
                {
                    "enabled": True,
                    "id": "1",
                    "params": {
                        "customLabel": "Median Indexing Throughput [docs/s]",
                        "field": "value.median",
                        "percents": [
                            50
                        ]
                    },
                    "schema": "metric",
                    "type": "median"
                },
                {
                    "enabled": True,
                    "id": "2",
                    "params": {
                        "field": "distribution-version",
                        "order": "asc",
                        "orderBy": "_term",
                        "size": 10
                    },
                    "schema": "segment",
                    "type": "terms"
                },
                {
                    "enabled": True,
                    "id": "3",
                    "params": {
                        "field": "user-tags.setup",
                        "order": "desc",
                        "orderBy": "_term",
                        "size": 10
                    },
                    "schema": "group",
                    "type": "terms"
                },
                {
                    "enabled": True,
                    "id": "4",
                    "params": {
                        "filters": filters
                    },
                    "schema": "split",
                    "type": "filters"
                }
            ],
            "listeners": {},
            "params": {
                "addLegend": True,
                "addTimeMarker": False,
                "addTooltip": True,
                "categoryAxes": [
                    {
                        "id": "CategoryAxis-1",
                        "labels": {
                            "show": True,
                            "truncate": 100
                        },
                        "position": "bottom",
                        "scale": {
                            "type": "linear"
                        },
                        "show": True,
                        "style": {},
                        "title": {
                            "text": "distribution-version: Ascending"
                        },
                        "type": "category"
                    }
                ],
                "defaultYExtents": False,
                "drawLinesBetweenPoints": True,
                "grid": {
                    "categoryLines": False,
                    "style": {
                        "color": "#eee"
                    }
                },
                "interpolate": "linear",
                "legendPosition": "right",
                "radiusRatio": 9,
                "scale": "linear",
                "seriesParams": [
                    {
                        "data": {
                            "id": "1",
                            "label": "Median Indexing Throughput [docs/s]"
                        },
                        "drawLinesBetweenPoints": True,
                        "mode": "normal",
                        "show": "True",
                        "showCircles": True,
                        "type": "histogram",
                        "valueAxis": "ValueAxis-1"
                    }
                ],
                "setYExtents": False,
                "showCircles": True,
                "times": [],
                "valueAxes": [
                    {
                        "id": "ValueAxis-1",
                        "labels": {
                            "filter": False,
                            "rotate": 0,
                            "show": True,
                            "truncate": 100
                        },
                        "name": "LeftAxis-1",
                        "position": "left",
                        "scale": {
                            "mode": "normal",
                            "type": "linear"
                        },
                        "show": True,
                        "style": {},
                        "title": {
                            "text": "Median Indexing Throughput [docs/s]"
                        },
                        "type": "value"
                    }
                ],
                "row": True
            },
            "title": title,
            "type": "histogram"
        }

        search_source = {
            "index": "benchmark-results-*",
            "query": {
                "query_string": {
                    "analyze_wildcard": True,
                    "query": "environment:\"%s\" AND active:true AND name:\"throughput\"" % environment
                }
            },
            "filter": []
        }

        return {
            "id": str(uuid.uuid4()),
            "type": "visualization",
            "attributes": {
                "title": title,
                "visState": json.dumps(vis_state),
                "uiStateJSON": BarCharts.UI_STATE_JSON,
                "description": "index",
                "version": 1,
                "kibanaSavedObjectMeta": {
                    "searchSourceJSON": json.dumps(search_source)
                }
            }
        }


class TimeSeriesCharts:
    @staticmethod
    def format_title(environment, workload_name, flavor=None, os_license=None, suffix=None):
        if flavor:
            title = [environment, flavor, str(workload_name)]
        elif os_license:
            title = [environment, os_license, str(workload_name)]
        elif flavor and os_license:
            raise exceptions.BenchmarkAssertionError(
                f"Specify either flavor [{flavor}] or license [{os_license}] but not both")
        else:
            title = [environment, str(workload_name)]
        if suffix:
            title.append(suffix)

        return "-".join(title)

    @staticmethod
    def filter_string(environment, test_ex_config):
        nightly_extra_filter = ""
        if test_ex_config.os_license:
            # Time series charts need to support different licenses and produce customized titles.
            nightly_extra_filter = f" AND user-tags.license:\"{test_ex_config.os_license}\""
        if test_ex_config.name:
            return f"environment:\"{environment}\" AND active:true AND user-tags.name:\"{test_ex_config.name}\"{nightly_extra_filter}"
        else:
            return f"environment:\"{environment}\" AND active:true AND workload:\"{test_ex_config.workload}\""\
                   f" AND test_procedure:\"{test_ex_config.test_procedure}\""\
                   f" AND provision_config_instance:\"{test_ex_config.provision_config_instance}\""\
                   f" AND node-count:{test_ex_config.node_count}"

    @staticmethod
    def gc(title, environment, test_ex_config):
        vis_state = {
            "title": title,
            "type": "metrics",
            "params": {
                "axis_formatter": "number",
                "axis_position": "left",
                "id": str(uuid.uuid4()),
                "index_pattern": "benchmark-results-*",
                "interval": "1d",
                "series": [
                    {
                        "axis_position": "left",
                        "chart_type": "line",
                        "color": "#68BC00",
                        "fill": "0",
                        "formatter": "number",
                        "id": str(uuid.uuid4()),
                        "line_width": "1",
                        "metrics": [
                            {
                                "id": str(uuid.uuid4()),
                                "type": "avg",
                                "field": "value.single"
                            }
                        ],
                        "point_size": "3",
                        "seperate_axis": 1,
                        "split_mode": "filters",
                        "stacked": "none",
                        "filter": "",
                        "split_filters": [
                            {
                                "filter": "young_gc_time",
                                "label": "Young Gen GC time",
                                "color": "rgba(0,191,179,1)",
                                "id": str(uuid.uuid4())
                            },
                            {
                                "filter": "old_gc_time",
                                "label": "Old Gen GC time",
                                "color": "rgba(254,209,10,1)",
                                "id": str(uuid.uuid4())
                            }
                        ],
                        "label": "GC Times",
                        "value_template": "{{value}} ms",
                        "steps": 0
                    }
                ],
                "show_legend": 1,
                "show_grid": 1,
                "drop_last_bucket": 0,
                "time_field": "test-execution-timestamp",
                "type": "timeseries",
                "filter": TimeSeriesCharts.filter_string(environment, test_ex_config),
                "annotations": [
                    {
                        "fields": "message",
                        "template": "{{message}}",
                        "index_pattern": "benchmark-annotations",
                        "query_string": f"((NOT _exists_:workload) OR workload:\"{test_ex_config.workload}\") "\
                            f"AND ((NOT _exists_:chart) OR chart:gc) "
                                        f"AND ((NOT _exists_:chart-name) OR chart-name:\"{title}\") AND environment:\"{environment}\"",
                        "id": str(uuid.uuid4()),
                        "color": "rgba(102,102,102,1)",
                        "time_field": "test-execution-timestamp",
                        "icon": "fa-tag",
                        "ignore_panel_filters": 1
                    }
                ],
                "axis_min": "0"
            },
            "aggs": [],
            "listeners": {}
        }

        return {
            "id": str(uuid.uuid4()),
            "type": "visualization",
            "attributes": {
                "title": title,
                "visState": json.dumps(vis_state),
                "uiStateJSON": "{}",
                "description": "gc",
                "version": 1,
                "kibanaSavedObjectMeta": {
                    "searchSourceJSON": "{\"query\":\"*\",\"filter\":[]}"
                }
            }
        }

    @staticmethod
    def merge_time(title, environment, test_execution_config):
        vis_state = {
            "title": title,
            "type": "metrics",
            "params": {
                "axis_formatter": "number",
                "axis_position": "left",
                "id": str(uuid.uuid4()),
                "index_pattern": "benchmark-results-*",
                "interval": "1d",
                "series": [
                    {
                        "axis_position": "left",
                        "chart_type": "line",
                        "color": "#68BC00",
                        "fill": "0",
                        "formatter": "number",
                        "id": str(uuid.uuid4()),
                        "line_width": "1",
                        "metrics": [
                            {
                                "id": str(uuid.uuid4()),
                                "type": "avg",
                                "field": "value.single"
                            }
                        ],
                        "point_size": "3",
                        "seperate_axis": 1,
                        "split_mode": "filters",
                        "stacked": "none",
                        "filter": "",
                        "split_filters": [
                            {
                                "filter": "merge_time",
                                "label": "Cumulative merge time",
                                "color": "rgba(0,191,179,1)",
                                "id": str(uuid.uuid4())
                            },
                            {
                                "filter": "merge_throttle_time",
                                "label": "Cumulative merge throttle time",
                                "color": "rgba(254,209,10,1)",
                                "id": str(uuid.uuid4())
                            }
                        ],
                        "label": "Merge Times",
                        "value_template": "{{value}} ms",
                        "steps": 0
                    }
                ],
                "show_legend": 1,
                "show_grid": 1,
                "drop_last_bucket": 0,
                "time_field": "test-execution-timestamp",
                "type": "timeseries",
                "filter": TimeSeriesCharts.filter_string(environment, test_execution_config),
                "annotations": [
                    {
                        "fields": "message",
                        "template": "{{message}}",
                        "index_pattern": "benchmark-annotations",
                        "query_string": f"((NOT _exists_:workload) OR workload:\"{test_execution_config.workload}\") "
                                        f"AND ((NOT _exists_:chart) OR chart:merge_times) "
                                        f"AND ((NOT _exists_:chart-name) OR chart-name:\"{title}\") AND environment:\"{environment}\"",
                        "id": str(uuid.uuid4()),
                        "color": "rgba(102,102,102,1)",
                        "time_field": "test-execution-timestamp",
                        "icon": "fa-tag",
                        "ignore_panel_filters": 1
                    }
                ],
                "axis_min": "0"
            },
            "aggs": [],
            "listeners": {}
        }

        return {
            "id": str(uuid.uuid4()),
            "type": "visualization",
            "attributes": {
                "title": title,
                "visState": json.dumps(vis_state),
                "uiStateJSON": "{}",
                "description": "merge_times",
                "version": 1,
                "kibanaSavedObjectMeta": {
                    "searchSourceJSON": "{\"query\":\"*\",\"filter\":[]}"
                }
            }
        }

    @staticmethod
    def merge_count(title, environment, test_execution_config):
        vis_state = {
            "title": title,
            "type": "metrics",
            "params": {
                "axis_formatter": "number",
                "axis_position": "left",
                "id": str(uuid.uuid4()),
                "index_pattern": "benchmark-results-*",
                "interval": "1d",
                "series": [
                    {
                        "axis_position": "left",
                        "chart_type": "line",
                        "color": "#68BC00",
                        "fill": "0",
                        "formatter": "number",
                        "id": str(uuid.uuid4()),
                        "line_width": "1",
                        "metrics": [
                            {
                                "id": str(uuid.uuid4()),
                                "type": "avg",
                                "field": "value.single"
                            }
                        ],
                        "point_size": "3",
                        "seperate_axis": 1,
                        "split_mode": "filters",
                        "stacked": "none",
                        "filter": "",
                        "split_filters": [
                            {
                                "filter": "merge_count",
                                "label": "Cumulative merge count",
                                "color": "rgba(0,191,179,1)",
                                "id": str(uuid.uuid4())
                            }
                        ],
                        "label": "Merge Count",
                        "value_template": "{{value}}",
                        "steps": 0
                    }
                ],
                "show_legend": 1,
                "show_grid": 1,
                "drop_last_bucket": 0,
                "time_field": "test-execution-timestamp",
                "type": "timeseries",
                "filter": TimeSeriesCharts.filter_string(environment, test_execution_config),
                "annotations": [
                    {
                        "fields": "message",
                        "template": "{{message}}",
                        "index_pattern": "benchmark-annotations",
                        "query_string": f"((NOT _exists_:workload) OR workload:\"{test_execution_config.workload}\") "
                                        f"AND ((NOT _exists_:chart) OR chart:merge_count) "
                                        f"AND ((NOT _exists_:chart-name) OR chart-name:\"{title}\") AND environment:\"{environment}\"",
                        "id": str(uuid.uuid4()),
                        "color": "rgba(102,102,102,1)",
                        "time_field": "test-execution-timestamp",
                        "icon": "fa-tag",
                        "ignore_panel_filters": 1
                    }
                ],
                "axis_min": "0"
            },
            "aggs": [],
            "listeners": {}
        }

        return {
            "id": str(uuid.uuid4()),
            "type": "visualization",
            "attributes": {
                "title": title,
                "visState": json.dumps(vis_state),
                "uiStateJSON": "{}",
                "description": "merge_count",
                "version": 1,
                "kibanaSavedObjectMeta": {
                    "searchSourceJSON": "{\"query\":\"*\",\"filter\":[]}"
                }
            }
        }

    @staticmethod
    def io(title, environment, test_ex_config):
        vis_state = {
            "title": title,
            "type": "metrics",
            "params": {
                "axis_formatter": "number",
                "axis_position": "left",
                "id": str(uuid.uuid4()),
                "index_pattern": "benchmark-results-*",
                "interval": "1d",
                "series": [
                    {
                        "axis_position": "left",
                        "chart_type": "line",
                        "color": "#68BC00",
                        "fill": "0",
                        "formatter": "bytes",
                        "id": str(uuid.uuid4()),
                        "line_width": "1",
                        "metrics": [
                            {
                                "id": str(uuid.uuid4()),
                                "type": "sum",
                                "field": "value.single"
                            }
                        ],
                        "point_size": "3",
                        "seperate_axis": 1,
                        "split_mode": "filters",
                        "stacked": "none",
                        "filter": "",
                        "split_filters": [
                            {
                                "filter": "name:index_size",
                                "label": "Index Size",
                                "color": "rgba(0,191,179,1)",
                                "id": str(uuid.uuid4())
                            },
                            {
                                "filter": "name:bytes_written",
                                "label": "Written",
                                "color": "rgba(254,209,10,1)",
                                "id": str(uuid.uuid4())
                            }
                        ],
                        "label": "Disk IO",
                        "value_template": "{{value}}",
                        "steps": 0
                    }
                ],
                "show_legend": 1,
                "show_grid": 1,
                "drop_last_bucket": 0,
                "time_field": "test-execution-timestamp",
                "type": "timeseries",
                "filter": TimeSeriesCharts.filter_string(environment, test_ex_config),
                "annotations": [
                    {
                        "fields": "message",
                        "template": "{{message}}",
                        "index_pattern": "benchmark-annotations",
                        "query_string": f"((NOT _exists_:workload) OR workload:\"{test_ex_config.workload}\") "\
                            f"AND ((NOT _exists_:chart) OR chart:io) "
                                        f"AND ((NOT _exists_:chart-name) OR chart-name:\"{title}\") AND environment:\"{environment}\"",
                        "id": str(uuid.uuid4()),
                        "color": "rgba(102,102,102,1)",
                        "time_field": "test-execution-timestamp",
                        "icon": "fa-tag",
                        "ignore_panel_filters": 1
                    }
                ],
                "axis_min": "0"
            },
            "aggs": [],
            "listeners": {}
        }

        return {
            "id": str(uuid.uuid4()),
            "type": "visualization",
            "attributes": {
                "title": title,
                "visState": json.dumps(vis_state),
                "uiStateJSON": "{}",
                "description": "io",
                "version": 1,
                "kibanaSavedObjectMeta": {
                    "searchSourceJSON": "{\"query\":\"*\",\"filter\":[]}"
                }
            }
        }

    @staticmethod
    def segment_memory(title, environment, test_ex_config):
        vis_state = {
            "title": title,
            "type": "metrics",
            "params": {
                "axis_formatter": "number",
                "axis_position": "left",
                "id": str(uuid.uuid4()),
                "index_pattern": "benchmark-results-*",
                "interval": "1d",
                "series": [
                    {
                        "axis_position": "left",
                        "chart_type": "line",
                        "color": "#68BC00",
                        "fill": "0",
                        "formatter": "bytes",
                        "id": str(uuid.uuid4()),
                        "line_width": "1",
                        "metrics": [
                            {
                                "id": str(uuid.uuid4()),
                                "type": "avg",
                                "field": "value.single"
                            }
                        ],
                        "point_size": "3",
                        "seperate_axis": 1,
                        "split_mode": "filters",
                        "stacked": "none",
                        "filter": f"environment:{environment} AND workload:\"{test_ex_config.workload}\"",
                        "split_filters": [
                            {
                                "filter": "memory_segments",
                                "label": "Segments",
                                "color": color_scheme_rgba[0],
                                "id": str(uuid.uuid4())
                            },
                            {
                                "filter": "memory_doc_values",
                                "label": "Doc Values",
                                "color": color_scheme_rgba[1],
                                "id": str(uuid.uuid4())
                            },
                            {
                                "filter": "memory_terms",
                                "label": "Terms",
                                "color": color_scheme_rgba[2],
                                "id": str(uuid.uuid4())
                            },
                            {
                                "filter": "memory_norms",
                                "label": "Norms",
                                "color": color_scheme_rgba[3],
                                "id": str(uuid.uuid4())
                            },
                            {
                                "filter": "memory_points",
                                "label": "Points",
                                "color": color_scheme_rgba[4],
                                "id": str(uuid.uuid4())
                            },
                            {
                                "filter": "memory_stored_fields",
                                "label": "Stored Fields",
                                "color": color_scheme_rgba[5],
                                "id": str(uuid.uuid4())
                            }
                        ],
                        "label": "Segment Memory",
                        "value_template": "{{value}}",
                        "steps": 0
                    }
                ],
                "show_legend": 1,
                "time_field": "test-execution-timestamp",
                "type": "timeseries",
                "filter": TimeSeriesCharts.filter_string(environment, test_ex_config),
                "annotations": [
                    {
                        "fields": "message",
                        "template": "{{message}}",
                        "index_pattern": "benchmark-annotations",
                        "query_string": f"((NOT _exists_:workload) OR workload:\"{test_ex_config.workload}\") "
                                        f"AND ((NOT _exists_:chart) OR chart:segment_memory) "
                                        f"AND ((NOT _exists_:chart-name) OR chart-name:\"{title}\") AND environment:\"{environment}\"",
                        "id": str(uuid.uuid4()),
                        "color": "rgba(102,102,102,1)",
                        "time_field": "test-execution-timestamp",
                        "icon": "fa-tag",
                        "ignore_panel_filters": 1
                    }
                ],
                "show_grid": 1,
                "drop_last_bucket": 0,
                "axis_min": "0"
            },
            "aggs": []
        }

        return {
            "id": str(uuid.uuid4()),
            "type": "visualization",
            "attributes": {
                "title": title,
                "visState": json.dumps(vis_state),
                "uiStateJSON": "{}",
                "description": "segment_memory",
                "version": 1,
                "kibanaSavedObjectMeta": {
                    "searchSourceJSON": "{\"query\":{\"query\":\"*\",\"language\":\"lucene\"},\"filter\":[]}"
                }
            }
        }

    @staticmethod
    def query(environment, test_ex_config, q):
        metric = "latency"
        title = TimeSeriesCharts.format_title(environment, test_ex_config.workload, os_license=test_ex_config.os_license,
                                              suffix="%s-%s-%s" % (test_ex_config.label, q, metric))

        vis_state = {
            "title": title,
            "type": "metrics",
            "params": {
                "id": str(uuid.uuid4()),
                "type": "timeseries",
                "series": [
                    {
                        "id": str(uuid.uuid4()),
                        "color": color_scheme_rgba[0],
                        "split_mode": "everything",
                        "label": "50th percentile",
                        "metrics": [
                            {
                                "id": str(uuid.uuid4()),
                                "type": "avg",
                                "field": "value.50_0"
                            }
                        ],
                        "seperate_axis": 0,
                        "axis_position": "right",
                        "formatter": "number",
                        "chart_type": "line",
                        "line_width": 1,
                        "point_size": 1,
                        "fill": "0.6",
                        "stacked": "none",
                        "split_color_mode": "gradient",
                        "series_drop_last_bucket": 0,
                        "value_template": "{{value}} ms",
                    },
                    {
                        "id": str(uuid.uuid4()),
                        "color": color_scheme_rgba[1],
                        "split_mode": "everything",
                        "label": "90th percentile",
                        "metrics": [
                            {
                                "id": str(uuid.uuid4()),
                                "type": "avg",
                                "field": "value.90_0"
                            }
                        ],
                        "seperate_axis": 0,
                        "axis_position": "right",
                        "formatter": "number",
                        "chart_type": "line",
                        "line_width": 1,
                        "point_size": 1,
                        "fill": "0.4",
                        "stacked": "none",
                        "split_color_mode": "gradient",
                        "series_drop_last_bucket": 0,
                        "value_template": "{{value}} ms",
                    },
                    {
                        "id": str(uuid.uuid4()),
                        "color": color_scheme_rgba[2],
                        "split_mode": "everything",
                        "label": "99th percentile",
                        "metrics": [
                            {
                                "id": str(uuid.uuid4()),
                                "type": "avg",
                                "field": "value.99_0"
                            }
                        ],
                        "seperate_axis": 0,
                        "axis_position": "right",
                        "formatter": "number",
                        "chart_type": "line",
                        "line_width": 1,
                        "point_size": 1,
                        "fill": "0.2",
                        "stacked": "none",
                        "split_color_mode": "gradient",
                        "series_drop_last_bucket": 0,
                        "value_template": "{{value}} ms",
                    },
                    {
                        "id": str(uuid.uuid4()),
                        "color": color_scheme_rgba[3],
                        "split_mode": "everything",
                        "label": "100th percentile",
                        "metrics": [
                            {
                                "id": str(uuid.uuid4()),
                                "type": "avg",
                                "field": "value.100_0"
                            }
                        ],
                        "seperate_axis": 0,
                        "axis_position": "right",
                        "formatter": "number",
                        "chart_type": "line",
                        "line_width": 1,
                        "point_size": 1,
                        "fill": "0.1",
                        "stacked": "none",
                        "split_color_mode": "gradient",
                        "series_drop_last_bucket": 0,
                        "value_template": "{{value}} ms",
                    }
                ],
                "time_field": "test-execution-timestamp",
                "index_pattern": "benchmark-results-*",
                "interval": "1d",
                "axis_position": "left",
                "axis_formatter": "number",
                "show_legend": 1,
                "show_grid": 1,
                "drop_last_bucket": 0,
                "background_color_rules": [
                    {
                        "id": str(uuid.uuid4())
                    }
                ],
                "filter": "task:\"%s\" AND name:\"%s\" AND %s" % (q, metric, TimeSeriesCharts.filter_string(
                    environment, test_ex_config)),
                "annotations": [
                    {
                        "fields": "message",
                        "template": "{{message}}",
                        "index_pattern": "benchmark-annotations",
                        "query_string": f"((NOT _exists_:workload) OR workload:\"{test_ex_config.workload}\") "
                                        f"AND ((NOT _exists_:chart) OR chart:query) "
                                        f"AND ((NOT _exists_:chart-name) OR chart-name:\"{title}\") AND environment:\"{environment}\"",
                        "id": str(uuid.uuid4()),
                        "color": "rgba(102,102,102,1)",
                        "time_field": "test-execution-timestamp",
                        "icon": "fa-tag",
                        "ignore_panel_filters": 1
                    }
                ]
            },
            "aggs": [],
            "listeners": {}
        }

        return {
            "id": str(uuid.uuid4()),
            "type": "visualization",
            "attributes": {
                "title": title,
                "visState": json.dumps(vis_state),
                "uiStateJSON": "{}",
                "description": "query",
                "version": 1,
                "kibanaSavedObjectMeta": {
                    "searchSourceJSON": "{\"query\":\"*\",\"filter\":[]}"
                }
            }
        }

    @staticmethod
    def index(environment, test_execution_configs, title):
        filters = []
        # any test_execution_config will do - they all belong to the same workload
        t = test_execution_configs[0].workload
        for idx, test_execution_config in enumerate(test_execution_configs):
            label = index_label(test_execution_config)
            for bulk_task in test_execution_config.bulk_tasks:
                filters.append(
                    {
                        "filter": "task:\"%s\" AND %s" % (bulk_task, TimeSeriesCharts.filter_string(environment, test_execution_config)),
                        "label": label,
                        "color": color_scheme_rgba[idx % len(color_scheme_rgba)],
                        "id": str(uuid.uuid4())
                    }
                )

        vis_state = {
            "title": title,
            "type": "metrics",
            "params": {
                "axis_formatter": "number",
                "axis_position": "left",
                "id": str(uuid.uuid4()),
                "index_pattern": "benchmark-results-*",
                "interval": "1d",
                "series": [
                    {
                        "axis_position": "left",
                        "chart_type": "line",
                        "color": "#68BC00",
                        "fill": "0",
                        "formatter": "number",
                        "id": str(uuid.uuid4()),
                        "line_width": "1",
                        "metrics": [
                            {
                                "id": str(uuid.uuid4()),
                                "type": "avg",
                                "field": "value.median"
                            }
                        ],
                        "point_size": "3",
                        "seperate_axis": 1,
                        "split_mode": "filters",
                        "stacked": "none",
                        "filter": "environment:\"%s\" AND workload:\"%s\"" % (environment, t),
                        "split_filters": filters,
                        "label": "Indexing Throughput",
                        "value_template": "{{value}} docs/s",
                        "steps": 0
                    }
                ],
                "show_legend": 1,
                "show_grid": 1,
                "drop_last_bucket": 0,
                "time_field": "test-execution-timestamp",
                "type": "timeseries",
                "filter": "environment:\"%s\" AND workload:\"%s\" AND name:\"throughput\" AND active:true" % (environment, t),
                "annotations": [
                    {
                        "fields": "message",
                        "template": "{{message}}",
                        "index_pattern": "benchmark-annotations",
                        "query_string": f"((NOT _exists_:workload) OR workload:\"{t}\") "
                                        f"AND ((NOT _exists_:chart) OR chart:indexing) "
                                        f"AND ((NOT _exists_:chart-name) OR chart-name:\"{title}\") AND environment:\"{environment}\"",
                        "id": str(uuid.uuid4()),
                        "color": "rgba(102,102,102,1)",
                        "time_field": "test-execution-timestamp",
                        "icon": "fa-tag",
                        "ignore_panel_filters": 1
                    }
                ],
                "axis_min": "0"
            },
            "aggs": [],
            "listeners": {}
        }
        return {
            "id": str(uuid.uuid4()),
            "type": "visualization",
            "attributes": {
                "title": title,
                "visState": json.dumps(vis_state),
                "uiStateJSON": "{}",
                "description": "index",
                "version": 1,
                "kibanaSavedObjectMeta": {
                    "searchSourceJSON": "{\"query\":\"*\",\"filter\":[]}"
                }
            }
        }


class TestExecutionConfigWorkload:
    def __init__(self, cfg, repository, name=None):
        self.repository = repository
        self.cached_workload = self.load_workload(cfg, name=name)

    def load_workload(self, cfg, name=None, params=None, excluded_tasks=None):
        if not params:
            params = {}
        # required in case a previous workload using a different repository has specified the revision
        if cfg.opts("workload", "repository.name", mandatory=False) != self.repository:
            cfg.add(config.Scope.applicationOverride, "workload", "repository.revision", None)
        # hack to make this work with multiple workloads (Benchmark core is usually not meant to be used this way)
        if name:
            cfg.add(config.Scope.applicationOverride, "workload", "repository.name", self.repository)
            cfg.add(config.Scope.applicationOverride, "workload", "workload.name", name)
        # another hack to ensure any workload-params in the test_execution config are used by Benchmark's workload loader
        cfg.add(config.Scope.applicationOverride, "workload", "params", params)
        if excluded_tasks:
            cfg.add(config.Scope.application, "workload", "exclude.tasks", excluded_tasks)
        return workload.load_workload(cfg)

    def get_workload(self, cfg, name=None, params=None, excluded_tasks=None):
        if params or excluded_tasks:
            return self.load_workload(cfg, name, params, excluded_tasks)
        # if no params specified, return the initially cached, (non-parametrized) workload
        return self.cached_workload


def generate_index_ops(chart_type, test_execution_configs, environment, logger):
    idx_test_execution_configs = list(filter(lambda c: "indexing" in c.charts, test_execution_configs))
    for test_execution_conf in idx_test_execution_configs:
        logger.debug("Gen index visualization for test_execution config with name:[%s] / label:[%s] / flavor: [%s] / license: [%s]",
                     test_execution_conf.name,
                     test_execution_conf.label,
                     test_execution_conf.flavor,
                     test_execution_conf.os_license)
    charts = []

    if idx_test_execution_configs:
        title = chart_type.format_title(
            environment,
            test_execution_configs[0].workload,
            flavor=test_execution_configs[0].flavor,
            suffix="indexing-throughput")
        charts = [chart_type.index(environment, idx_test_execution_configs, title)]
    return charts


def generate_queries(chart_type, test_execution_configs, environment):
    # output JSON structures
    structures = []

    for test_execution_config in test_execution_configs:
        if "query" in test_execution_config.charts:
            for q in test_execution_config.throttled_tasks:
                structures.append(chart_type.query(environment, test_execution_config, q))
    return structures


def generate_io(chart_type, test_execution_configs, environment):
    # output JSON structures
    structures = []
    for test_execution_config in test_execution_configs:
        if "io" in test_execution_config.charts:
            title = chart_type.format_title(environment, test_execution_config.workload, os_license=test_execution_config.os_license,
                                            suffix="%s-io" % test_execution_config.label)
            structures.append(chart_type.io(title, environment, test_execution_config))

    return structures


def generate_gc(chart_type, test_execution_configs, environment):
    structures = []
    for test_execution_config in test_execution_configs:
        if "gc" in test_execution_config.charts:
            title = chart_type.format_title(environment, test_execution_config.workload, os_license=test_execution_config.os_license,
                                            suffix="%s-gc" % test_execution_config.label)
            structures.append(chart_type.gc(title, environment, test_execution_config))

    return structures

def generate_merge_time(chart_type, test_execution_configs, environment):
    structures = []
    for test_execution_config in test_execution_configs:
        if "merge_times" in test_execution_config.charts:
            title = chart_type.format_title(environment, test_execution_config.workload, os_license=test_execution_config.os_license,
                                            suffix=f"{test_execution_config.label}-merge-times")
            structures.append(chart_type.merge_time(title, environment, test_execution_config))

    return structures

def generate_merge_count(chart_type, test_execution_configs, environment):
    structures = []
    for test_execution_config in test_execution_configs:
        if "merge_count" in test_execution_config.charts:
            title = chart_type.format_title(environment, test_execution_config.workload, os_license=test_execution_config.os_license,
                                            suffix=f"{test_execution_config.label}-merge-count")
            structures.append(chart_type.merge_count(title, environment, test_execution_config))

    return structures


def generate_segment_memory(chart_type, test_execution_configs, environment):
    structures = []
    for test_execution_config in test_execution_configs:
        if "segment_memory" in test_execution_config.charts:
            title = chart_type.format_title(environment, test_execution_config.workload, os_license=test_execution_config.os_license,
                                            suffix="%s-segment-memory" % test_execution_config.label)
            chart = chart_type.segment_memory(title, environment, test_execution_config)
            if chart:
                structures.append(chart)
    return structures


def generate_dashboard(chart_type, environment, workload, charts, flavor=None):
    panels = []

    width = 24
    height = 32

    row = 0
    col = 0

    for idx, chart in enumerate(charts):
        panelIndex = idx + 1
        # make index charts wider
        if chart["attributes"]["description"] == "index":
            chart_width = 2 * width
            # force one panel per row
            next_col = 0
        else:
            chart_width = width
            # two rows per panel
            next_col = (col + 1) % 2

        panel = {
            "id": chart["id"],
            "panelIndex": panelIndex,
            "gridData": {
                "x": (col * chart_width),
                "y": (row * height),
                "w": chart_width,
                "h": height,
                "i": str(panelIndex)
            },
            "type": "visualization",
            "version": "7.10.2"
        }
        panels.append(panel)
        col = next_col
        if col == 0:
            row += 1

    return {
        "id": str(uuid.uuid4()),
        "type": "dashboard",
        "attributes": {
            "title": chart_type.format_title(environment, workload.name, flavor=flavor),
            "hits": 0,
            "description": "",
            "panelsJSON": json.dumps(panels),
            "optionsJSON": "{\"darkTheme\":false}",
            "uiStateJSON": "{}",
            "version": 1,
            "timeRestore": False,
            "kibanaSavedObjectMeta": {
                "searchSourceJSON": json.dumps(
                    {
                        "filter": [
                            {
                                "query": {
                                    "query_string": {
                                        "analyze_wildcard": True,
                                        "query": "*"
                                    }
                                }
                            }
                        ],
                        "highlightAll": True,
                        "version": True
                    }
                )
            }
        }
    }


class TestExecutionConfig:
    def __init__(self, workload, cfg=None, flavor=None, os_license=None, \
        test_procedure=None, provision_config_instance=None, node_count=None,\
             charts=None):
        self.workload = workload
        if cfg:
            self.configuration = cfg
            self.configuration["flavor"] = flavor
            self.configuration["os_license"] = os_license
        else:
            self.configuration = {
                "charts": charts,
                "test_procedure": test_procedure,
                "provision-config-instance": provision_config_instance,
                "node-count": node_count
            }

    @property
    def name(self):
        return self.configuration.get("name")

    @property
    def flavor(self):
        return self.configuration.get("flavor")

    @property
    def os_license(self):
        return self.configuration.get("os_license")

    @property
    def label(self):
        return self.configuration.get("label")

    @property
    def charts(self):
        return self.configuration["charts"]

    @property
    def node_count(self):
        return self.configuration.get("node-count", 1)

    @property
    def test_procedure(self):
        return self.configuration["test_procedure"]

    @property
    def provision_config_instance(self):
        return self.configuration["provision-config-instance"]

    @property
    def plugins(self):
        return self.configuration.get("plugins", "")

    @property
    def bulk_tasks(self):
        task_names = []
        for task in self.workload.find_test_procedure_or_default(self.test_procedure).schedule:
            for sub_task in task:
                # We are looking for type bulk operations to add to indexing throughput chart.
                # For the observability workload, the index operation is of type raw-bulk, instead of type bulk.
                # Doing a lenient match to allow for that.
                if workload.OperationType.Bulk.to_hyphenated_string() in sub_task.operation.type:
                    if workload.OperationType.Bulk.to_hyphenated_string() != sub_task.operation.type:
                        console.info(f"Found [{sub_task.name}] of type [{sub_task.operation.type}] in "\
                                     f"[{self.test_procedure}], adding it to indexing dashboard.\n", flush=True)
                    task_names.append(sub_task.name)
        return task_names

    @property
    def throttled_tasks(self):
        task_names = []
        for task in self.workload.find_test_procedure_or_default(self.test_procedure).schedule:
            for sub_task in task:
                # We are assuming here that each task with a target throughput or target interval is interesting for latency charts.
                # We should refactor the chart generator to make this classification logic more flexible so the user can specify
                # which tasks / or types of operations should be used for which chart types.
                if "target-throughput" in sub_task.params or "target-interval" in sub_task.params:
                    task_names.append(sub_task.name)
        return task_names


def load_test_execution_configs(cfg, chart_type, chart_spec_path=None):
    def add_configs(test_execution_configs_per_lic, flavor_name="oss", lic="oss", workload_name=None):
        configs_per_lic = []
        for test_execution_config in test_execution_configs_per_lic:
            excluded_tasks = None
            if "exclude-tasks" in test_execution_config:
                excluded_tasks = test_execution_config.get("exclude-tasks").split(",")
            configs_per_lic.append(
                TestExecutionConfig(workload=test_execution_config_workload.get_workload(cfg, name=workload_name,
                                                             params=test_execution_config.get("workload-params", {}),
                                                             excluded_tasks=excluded_tasks),
                           cfg=test_execution_config,
                           flavor=flavor_name,
                           os_license=lic)
            )
        return configs_per_lic

    def add_test_execution_configs(license_configs, flavor_name, workload_name):
        if chart_type == BarCharts:
            # Only one license config, "basic", is present in bar charts
            _lic_conf = [license_config["configurations"] for license_config in license_configs if license_config["name"] == "basic"]
            if _lic_conf:
                test_execution_configs_per_workload.extend(add_configs(_lic_conf[0], workload_name=workload_name))
        else:
            for lic_config in license_configs:
                test_execution_configs_per_workload.extend(add_configs(lic_config["configurations"],
                                                          flavor_name,
                                                          lic_config["name"],
                                                          workload_name))

    test_execution_configs = {"oss": [], "default": []}
    if chart_type == BarCharts:
        test_execution_configs = []

    for _workload_file in glob.glob(io.normalize_path(chart_spec_path)):
        with open(_workload_file, mode="rt", encoding="utf-8") as f:
            for item in json.load(f):
                _workload_repository = item.get("workload-repository", "default")
                test_execution_config_workload = TestExecutionConfigWorkload(cfg, _workload_repository, name=item["workload"])
                for flavor in item["flavors"]:
                    test_execution_configs_per_workload = []
                    _flavor_name = flavor["name"]
                    _workload_name = item["workload"]
                    add_test_execution_configs(flavor["licenses"], _flavor_name, _workload_name)

                    if test_execution_configs_per_workload:
                        if chart_type == BarCharts:
                            test_execution_configs.append(test_execution_configs_per_workload)
                        else:
                            test_execution_configs[_flavor_name].append(test_execution_configs_per_workload)
    return test_execution_configs


def gen_charts_per_workload_configs(test_execution_configs, chart_type, env, flavor=None, logger=None):
    charts = generate_index_ops(chart_type, test_execution_configs, env, logger) + \
             generate_io(chart_type, test_execution_configs, env) + \
             generate_gc(chart_type, test_execution_configs, env) + \
             generate_merge_time(chart_type, test_execution_configs, env) + \
             generate_merge_count(chart_type, test_execution_configs, env) + \
             generate_segment_memory(chart_type, test_execution_configs, env) + \
             generate_queries(chart_type, test_execution_configs, env)

    dashboard = generate_dashboard(chart_type, env, test_execution_configs[0].workload, charts, flavor)

    return charts, dashboard


def gen_charts_per_workload(test_execution_configs, chart_type, env, flavor=None, logger=None):
    structures = []
    for test_execution_configs_per_workload in test_execution_configs:
        charts, dashboard = gen_charts_per_workload_configs(test_execution_configs_per_workload, chart_type, env, flavor, logger)
        structures.extend(charts)
        structures.append(dashboard)

    return structures


def gen_charts_from_workload_combinations(test_execution_configs, chart_type, env, logger):
    structures = []
    for flavor, test_execution_configs_per_flavor in test_execution_configs.items():
        for test_execution_configs_per_workload in test_execution_configs_per_flavor:
            logger.debug("Generating charts for test_execution_configs with name:[%s]/flavor:[%s]",
                         test_execution_configs_per_workload[0].name, flavor)
            charts, dashboard = gen_charts_per_workload_configs(test_execution_configs_per_workload, chart_type, env, flavor, logger)

            structures.extend(charts)
            structures.append(dashboard)

    return structures


def generate(cfg):
    logger = logging.getLogger(__name__)

    chart_spec_path = cfg.opts("generator", "chart.spec.path")
    if cfg.opts("generator", "chart.type") == "time-series":
        chart_type = TimeSeriesCharts
    else:
        chart_type = BarCharts

    console.info("Loading workload data...", flush=True)
    test_execution_configs = load_test_execution_configs(cfg, chart_type, chart_spec_path)
    env = cfg.opts("system", "env.name")

    structures = []
    console.info("Generating charts...", flush=True)

    if chart_type == BarCharts:
        # bar charts are flavor agnostic and split results based on a separate `user.setup` field
        structures = gen_charts_per_workload(test_execution_configs, chart_type, env, logger=logger)
    elif chart_type == TimeSeriesCharts:
        structures = gen_charts_from_workload_combinations(test_execution_configs, chart_type, env, logger)

    output_path = cfg.opts("generator", "output.path")
    if output_path:
        with open(io.normalize_path(output_path), mode="wt", encoding="utf-8") as f:
            for record in structures:
                print(json.dumps(record), file=f)
    else:
        for record in structures:
            print(json.dumps(record))



# SPDX-License-Identifier: Apache-2.0
#
# The OpenSearch Contributors require contributions made to
# this file be licensed under the Apache-2.0 license or a
# compatible open source license.
# Modifications Copyright OpenSearch Contributors. See
# GitHub history for details.
# Licensed to Elasticsearch B.V. under one or more contributor
# license agreements. See the NOTICE file distributed with
# this work for additional information regarding copyright
# ownership. Elasticsearch B.V. licenses this file to you under
# the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#	http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import logging
import os

from opensearchpy import OpenSearchException
from jinja2 import Environment, FileSystemLoader, select_autoescape

from osbenchmark import PROGRAM_NAME
from osbenchmark.client import OsClientFactory
from osbenchmark.workload_generator import corpus, index
from osbenchmark.utils import io, opts, console


def process_template(templates_path, template_filename, template_vars, output_path):
    env = Environment(loader=FileSystemLoader(templates_path), autoescape=select_autoescape(['html', 'xml']))
    template = env.get_template(template_filename)

    with open(output_path, "w") as f:
        f.write(template.render(template_vars))


def extract_mappings_and_corpora(client, output_path, indices_to_extract):
    indices = []
    corpora = []
    # first extract index metadata (which is cheap) and defer extracting data to reduce the potential for
    # errors due to invalid index names late in the process.
    for index_name in indices_to_extract:
        try:
            indices += index.extract(client, output_path, index_name)
        except OpenSearchException:
            logging.getLogger(__name__).exception("Failed to extract index [%s]", index_name)

    # That list only contains valid indices (with index patterns already resolved)
    for i in indices:
        c = corpus.extract(client, output_path, i["name"])
        if c:
            corpora.append(c)

    return indices, corpora


def create_workload(cfg):
    logger = logging.getLogger(__name__)

    workload_name = cfg.opts("workload", "workload.name")
    indices = cfg.opts("generator", "indices")
    root_path = cfg.opts("generator", "output.path")
    target_hosts = cfg.opts("client", "hosts")
    client_options = cfg.opts("client", "options")

    logger.info("Creating workload [%s] matching indices [%s]", workload_name, indices)

    client = OsClientFactory(hosts=target_hosts.all_hosts[opts.TargetHosts.DEFAULT],
                             client_options=client_options.all_client_options[opts.TargetHosts.DEFAULT]).create()

    info = client.info()
    console.info(f"Connected to OpenSearch cluster [{info['name']}] version [{info['version']['number']}].\n", logger=logger)

    output_path = os.path.abspath(os.path.join(io.normalize_path(root_path), workload_name))
    io.ensure_dir(output_path)

    indices, corpora = extract_mappings_and_corpora(client, output_path, indices)
    if len(indices) == 0:
        raise RuntimeError("Failed to extract any indices for workload!")

    template_vars = {
        "workload_name": workload_name,
        "indices": indices,
        "corpora": corpora
    }

    workload_path = os.path.join(output_path, "workload.json")
    templates_path = os.path.join(cfg.opts("node", "benchmark.root"), "resources")
    process_template(templates_path, "workload.json.j2", template_vars, workload_path)

    console.println("")
    console.info(f"Workload {workload_name} has been created. Run it with: {PROGRAM_NAME} --workload-path={output_path}")



# SPDX-License-Identifier: Apache-2.0
#
# The OpenSearch Contributors require contributions made to
# this file be licensed under the Apache-2.0 license or a
# compatible open source license.
# Modifications Copyright OpenSearch Contributors. See
# GitHub history for details.
# Licensed to Elasticsearch B.V. under one or more contributor
# license agreements. See the NOTICE file distributed with
# this work for additional information regarding copyright
# ownership. Elasticsearch B.V. licenses this file to you under
# the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#	http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import logging
import os
import sys

from osbenchmark import exceptions, config
from osbenchmark.utils import io, git, console, versions


class BenchmarkRepository:
    """
    Manages Benchmark resources (e.g. provision_configs or workloads).
    """

    default = "default-provision-config"

    def __init__(self, default_directory, root_dir, repo_name, resource_name, offline, fetch=True):
        # If no URL is found, we consider this a local only repo (but still require that it is a git repo)
        self.directory = default_directory
        self.remote = self.directory is not None and self.directory != BenchmarkRepository.default and self.directory.strip() != ""
        self.repo_dir = os.path.join(root_dir, repo_name)
        self.resource_name = resource_name
        self.offline = offline
        self.logger = logging.getLogger(__name__)
        self.revision = None

        if self.remote and not self.offline and fetch:
            # a normal git repo with a remote
            if not git.is_working_copy(self.repo_dir):
                git.clone(src=self.repo_dir, remote=self.directory)
            else:
                try:
                    git.fetch(src=self.repo_dir)
                except exceptions.SupplyError:
                    console.warn("Could not update %s. Continuing with your locally available state." % self.resource_name)
        else:
            if not git.is_working_copy(self.repo_dir) and repo_name != "default":
                if io.exists(self.repo_dir):
                    raise exceptions.SystemSetupError("[{src}] must be a git repository.\n\nPlease run:\ngit -C {src} init"
                                                      .format(src=self.repo_dir))

    def update(self, distribution_version):
        try:
            if self.remote:
                branch = versions.best_match(git.branches(self.repo_dir, remote=self.remote), distribution_version)
                if branch:
                    # Allow uncommitted changes iff we do not have to change the branch
                    self.logger.info(
                        "Checking out [%s] in [%s] for distribution version [%s].", branch, self.repo_dir, distribution_version)
                    git.checkout(self.repo_dir, branch=branch)
                    self.logger.info("Rebasing on [%s] in [%s] for distribution version [%s].", branch, self.repo_dir, distribution_version)
                    try:
                        git.rebase(self.repo_dir, branch=branch)
                        self.revision = git.head_revision(self.repo_dir)
                    except exceptions.SupplyError:
                        self.logger.exception("Cannot rebase due to local changes in [%s]", self.repo_dir)
                        console.warn(
                            "Local changes in [%s] prevent %s update from remote. Please commit your changes." %
                            (self.repo_dir, self.resource_name))
                    return
                else:
                    msg = "Could not find %s remotely for distribution version [%s]. Trying to find %s locally." % \
                          (self.resource_name, distribution_version, self.resource_name)
                    self.logger.warning(msg)
            branch = versions.best_match(git.branches(self.repo_dir, remote=False), distribution_version)
            if branch:
                if git.current_branch(self.repo_dir) != branch:
                    self.logger.info("Checking out [%s] in [%s] for distribution version [%s].",
                                     branch, self.repo_dir, distribution_version)
                    git.checkout(self.repo_dir, branch=branch)
                    self.revision = git.head_revision(self.repo_dir)
            else:
                self.logger.info("No local branch found for distribution version [%s] in [%s]. Checking tags.",
                                 distribution_version, self.repo_dir)
                tag = self._find_matching_tag(distribution_version)
                if tag:
                    self.logger.info("Checking out tag [%s] in [%s] for distribution version [%s].",
                                     tag, self.repo_dir, distribution_version)
                    git.checkout(self.repo_dir, branch=tag)
                    self.revision = git.head_revision(self.repo_dir)
                else:
                    raise exceptions.SystemSetupError("Cannot find %s for distribution version %s"
                                                      % (self.resource_name, distribution_version))
        except exceptions.SupplyError as e:
            tb = sys.exc_info()[2]
            raise exceptions.DataError("Cannot update %s in [%s] (%s)." % (self.resource_name, self.repo_dir, e.message)).with_traceback(tb)

    def _find_matching_tag(self, distribution_version):
        tags = git.tags(self.repo_dir)
        for version in versions.variants_of(distribution_version):
            # tags have a "v" prefix by convention.
            tag_candidate = "v{}".format(version)
            if tag_candidate in tags:
                return tag_candidate
        return None

    def checkout(self, revision):
        self.logger.info("Checking out revision [%s] in [%s].", revision, self.repo_dir)
        git.checkout(self.repo_dir, revision)

    def set_provision_configs_dir(self, repo_revision, distribution_version, cfg):
        if self.directory == BenchmarkRepository.default:
            self._use_default_provision_configs_dir(distribution_version, cfg)
        elif repo_revision:
            self.checkout(repo_revision)
        else:
            self.update(distribution_version)
            cfg.add(config.Scope.applicationOverride, "builder", "repository.revision", self.revision)

        return cfg

    def _use_default_provision_configs_dir(self, distribution_version, cfg):
        self.logger.info("Using default-provision-config directory within repository")
        root_dir = cfg.opts("node", "benchmark.root")
        provision_configs_path = os.path.join(root_dir, "resources/provision_configs")

        branch_version = self._select_branch_version(distribution_version, provision_configs_path)

        # Include selected branch
        self.repo_dir = os.path.join(provision_configs_path, branch_version)

    def _select_branch_version(self, distribution_version, pc_path):
        # Branches have been moved into resources/provision_configs
        branches = [b for b in os.listdir(pc_path) if os.path.isdir(os.path.join(pc_path, b)) and b != "main"]
        branches.sort(key=lambda b: list(map(int, b.split('.'))), reverse=True)
        self.logger.info("branches: %s", branches)
        convert = lambda s: list(map(int, s.split('.')))
        if distribution_version is not None:
            # Return a branch that is less than or equal to the distribution version
            for branch in branches:
                if convert(distribution_version) >= convert(branch):
                    return branch

            raise Exception("Distribution version is less than available ES versions for provision-configs.")

        # Distribution version is Nonetype if not specified in command line
        return "main"



# SPDX-License-Identifier: Apache-2.0
#
# The OpenSearch Contributors require contributions made to
# this file be licensed under the Apache-2.0 license or a
# compatible open source license.
# Modifications Copyright OpenSearch Contributors. See
# GitHub history for details.
# Licensed to Elasticsearch B.V. under one or more contributor
# license agreements. See the NOTICE file distributed with
# this work for additional information regarding copyright
# ownership. Elasticsearch B.V. licenses this file to you under
# the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#	http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import time
from datetime import datetime, timezone


def to_epoch_millis(t):
    """
    Convert a time instance retrieved via time.time() to a timestamp in milliseconds since epoch.
    :param t: a time instance
    :return: the corresponding value of milliseconds since epoch.
    """
    return int(round(t * 1000))


def to_iso8601(dt):
    """
    Convert a datetime instance to a ISO-8601 compliant string.
    :param dt: A datetime instance
    :return: The corresponding ISO-8601 formatted string
    """
    return "%04d%02d%02dT%02d%02d%02dZ" % (dt.year, dt.month, dt.day, dt.hour, dt.minute, dt.second)


def from_is8601(ts):
    """
    Convert an ISO-8601 compliant string (as created by ``to_iso8601(dt)``) to a datetime instance.

    :param ts: an ISO-8601 compliant string
    :return: The corresponding datetime instance.
    """
    return datetime.strptime(ts, "%Y%m%dT%H%M%SZ")


def sleep(seconds):
    time.sleep(seconds)


def _to_datetime(val, date_format=None, default_tz=None):
    if isinstance(val, datetime):
        return val
    # unix timestamp
    elif isinstance(val, float):
        return datetime.fromtimestamp(val, default_tz)
    elif isinstance(val, str):
        ret = datetime.strptime(val, date_format)
        if ret.tzinfo is None:
            ret = ret.replace(tzinfo=default_tz)
        return ret
    else:
        raise TypeError("Cannot convert unrecognized type '%s' with value '%s' to datetime." % (type(val), str(val)))


def days_ago(start_date, end_date, date_format="%d-%m-%Y", default_tz=timezone.utc):
    """

    Calculates the difference in days between a start date and an end date.

    :param start_date: The start date. May be a datetime instance, a unix timestamp or a string representation of a date.
    :param end_date: The end date. May be a datetime instance, a unix timestamp or a string representation of a date.
    :param date_format: If one or both date values are provided as strings, the date format which is needed for conversion.
    Default: "%d-%m-%Y"
    :param default_tz: Timezone to use for dates using unix timestamp or not already specified in date_format
    Default: timezone.UTC
    :return: The difference between start and end date in complete days.
    """
    start = _to_datetime(start_date, date_format, default_tz)
    end = _to_datetime(end_date, date_format, default_tz)
    return (end - start).days


class Clock:
    """
    A clock abstracts time measurements. Its main purpose is to ease testing
    """

    @staticmethod
    def now():
        """
        :return: The current time.
        """
        return time.time()

    @staticmethod
    def stop_watch():
        return StopWatch()


class StopWatch:
    def __init__(self):
        self._start = None
        self._stop = None

    def start(self):
        self._start = self._now()

    def stop(self):
        self._stop = self._now()

    def split_time(self):
        return self._interval(self._start, self._now())

    def total_time(self):
        return self._interval(self._start, self._stop)

    def _interval(self, t0, t1):
        if t0 is None:
            raise RuntimeError("start time is None")
        if t1 is None:
            raise RuntimeError("end time is None")
        return t1 - t0

    def _now(self):
        return time.perf_counter()



# SPDX-License-Identifier: Apache-2.0
#
# The OpenSearch Contributors require contributions made to
# this file be licensed under the Apache-2.0 license or a
# compatible open source license.
# Modifications Copyright OpenSearch Contributors. See
# GitHub history for details.
# Licensed to Elasticsearch B.V. under one or more contributor
# license agreements. See the NOTICE file distributed with
# this work for additional information regarding copyright
# ownership. Elasticsearch B.V. licenses this file to you under
# the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#	http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import re
import functools

from osbenchmark import exceptions

VERSIONS = re.compile(r"^(\d+)\.(\d+)\.(\d+)(?:-(.+))?$")

VERSIONS_OPTIONAL = re.compile(r"^(\d+)(?:\.(\d+))?(?:\.(\d+))?(?:-(.+))?$")


def _versions_pattern(strict):
    return VERSIONS if strict else VERSIONS_OPTIONAL


def is_version_identifier(text, strict=True):
    return text is not None and _versions_pattern(strict).match(text) is not None


def major_version(version):
    """
    Determines the major version of a given version string.

    :param version: A version string in the format major.minor.path-suffix (suffix is optional)
    :return: The major version (as int). In case the version string is invalid, an ``exceptions.InvalidSyntax`` is raised.
    """
    major, _, _, _ = components(version)
    return major


def components(version, strict=True):
    """
    Determines components of a version string.

    :param version: A version string in the format major.minor.path-suffix (suffix is optional)
    :param strict: Determines whether versions need to have at least "major", "minor" and "patch" defined. Default: True
    :return: A tuple with four components determining "major", "minor", "patch" and "suffix" (any part except "major" may be `None`)
    """
    versions_pattern = _versions_pattern(strict)
    matches = versions_pattern.match(version)
    if matches:
        if matches.start(4) > 0:
            return int(matches.group(1)), int(matches.group(2)), int(matches.group(3)), matches.group(4)
        elif matches.start(3) > 0:
            return int(matches.group(1)), int(matches.group(2)), int(matches.group(3)), None
        elif matches.start(2) > 0:
            return int(matches.group(1)), int(matches.group(2)), None, None
        elif matches.start(1) > 0:
            return int(matches.group(1)), None, None, None
        else:
            return int(version), None, None, None
    raise exceptions.InvalidSyntax("version string '%s' does not conform to pattern '%s'" % (version, versions_pattern.pattern))


@functools.total_ordering
class Version:
    """
    Represents a version with components major, minor, patch and suffix (suffix is optional). Suffixes are not
    considered for version comparisons as its contents are opaque and a semantically correct order cannot be defined.
    """
    def __init__(self, major, minor, patch, suffix=None):
        self.major = major
        self.minor = minor
        self.patch = patch
        self.suffix = suffix

    def __eq__(self, o: object) -> bool:
        return isinstance(o, type(self)) and (self.major, self.minor, self.patch) == (o.major, o.minor, o.patch)

    def __lt__(self, o: object) -> bool:
        return isinstance(o, type(self)) and (self.major, self.minor, self.patch) < (o.major, o.minor, o.patch)

    def __hash__(self) -> int:
        return hash(self.major) ^ hash(self.minor) ^ hash(self.patch) ^ hash(self.suffix)

    def __repr__(self) -> str:
        v = f"{self.major}.{self.minor}.{self.patch}"
        return f"{v}-{self.suffix}" if self.suffix else v

    @classmethod
    def from_string(cls, v):
        return cls(*components(v))


def variants_of(version):
    for v, _ in VersionVariants(version).all_versions:
        yield v


class VersionVariants:
    """
    Build all possible variations of a version.

    e.g. if version "5.0.0-SNAPSHOT" is given, the possible variations are:
        self.with_suffix: "5.0.0-SNAPSHOT",
        self.with_patch: "5.0.0",
        self.with_minor: "5.0",
        self.with_major: "5"
    """

    def __init__(self, version):
        """
        :param version: A version string in the format major.minor.path-suffix (suffix is optional)
        """
        self.major, self.minor, self.patch, self.suffix = components(version)

        self.with_major = f"{int(self.major)}"
        self.with_minor = f"{int(self.major)}.{int(self.minor)}"
        self.with_patch = f"{int(self.major)}.{int(self.minor)}.{int(self.patch)}"
        self.with_suffix = f"{int(self.major)}.{int(self.minor)}.{int(self.patch)}-{self.suffix}" if self.suffix \
            else None

    @property
    def all_versions(self):
        """
        :return: a list of tuples containing version variants and version type
            ordered from most specific to most generic variation.

            Example:
            [("5.0.0-SNAPSHOT", "with_suffix"), ("5.0.0", "with_patch"), ("5.0", "with_minor"), ("5", "with_major")]
        """

        versions = [(self.with_suffix, "with_suffix")] if self.suffix else []
        versions.extend([(self.with_patch, "with_patch"),
                         (self.with_minor, "with_minor"),
                         (self.with_major, "with_major")])

        return versions


def best_match(available_alternatives, distribution_version):
    """
    Finds the most specific branch for a given distribution version assuming that versions have the pattern:

        major.minor.patch-suffix

    and the provided alternatives reflect this pattern.
    Best matches for distribution_version from available_alternatives may be:
     1. exact matches of major.minor
     2. nearest prior minor within the same major
     3. major version
     4. as a last resort, `main`.

    See test_find_best_match() for examples.

    :param available_alternatives: A list of possible distribution versions (or shortened versions).
    :param distribution_version: An OpenSearch distribution version.
    :return: The most specific alternative that is available or None.
    """
    if is_version_identifier(distribution_version):
        versions = VersionVariants(distribution_version)
        for version, version_type in versions.all_versions:
            if version in available_alternatives:
                return version
            # match nearest prior minor
            if version_type == "with_minor" and (latest_minor := latest_bounded_minor(available_alternatives, versions)):
                if latest_minor:
                    return f"{versions.major}.{latest_minor}"
        # not found in the available alternatives, it could still be a main version
        major, _, _, _ = components(distribution_version)
        if major > _latest_major(available_alternatives):
            return "main"
    elif not distribution_version:
        return "main"
    return None


def _latest_major(alternatives):
    max_major = -1
    for a in alternatives:
        if is_version_identifier(a, strict=False):
            major, _, _, _ = components(a, strict=False)
            max_major = max(major, max_major)
    return max_major


def latest_bounded_minor(alternatives, target_version):
    """
    Finds the closest minor version that is smaller or eq to target_version from a list of version alternatives.
    Versions including patch or patch-suffix in alternatives are ignored.
    See test_latest_bounded_minor() for examples.

    :param alternatives: list of alternative versions
    :param target_version: a VersionVariants object presenting the distribution version
    :return: the closest minor version (if available) from alternatives otherwise None
    """
    eligible_minors = []
    for a in alternatives:
        if is_version_identifier(a, strict=False):
            major, minor, patch, suffix = components(a, strict=False)
            if patch is not None or suffix is not None:
                # branches containing patch or patch-suffix aren't supported
                continue
            if major == target_version.major and minor and minor <= target_version.minor:
                eligible_minors.append(minor)

    # no matching minor version
    if not eligible_minors:
        return None

    eligible_minors.sort()

    return min(eligible_minors, key=lambda x: abs(x - target_version.minor))



from abc import ABC, abstractmethod


class Installer(ABC):
    """
    Installers are invoked to prepare the OpenSearch and Plugin data that exists on a host so that an OpenSearch cluster
    can be started.
    """

    def __init__(self, executor):
        self.executor = executor

    @abstractmethod
    def install(self, host, binaries, all_node_ips):
        """
        Executes the necessary logic to prepare and install OpenSearch and any request Plugins on a cluster host

        ;param host: A Host object defining the host on which to install the data
        ;param binaries: A map of components to install to their paths on the host
        ;param all_node_ips: A list of the ips for each node in the cluster. Used for cluster formation
        ;return node: A Node object detailing the installation data of the node on the host
        """
        raise NotImplementedError

    @abstractmethod
    def cleanup(self, host):
        """
        Removes the data that was downloaded, installed, and created on a given host during the test execution

        ;param host: A Host object defining the host on which to remove the data
        ;return None
        """
        raise NotImplementedError



import logging

from osbenchmark.utils import console


class HostCleaner:
    def __init__(self, path_manager):
        self.logger = logging.getLogger(__name__)
        self.path_manager = path_manager

    def cleanup(self, host, preserve_install):
        if preserve_install:
            console.info("Preserving benchmark candidate installation.", logger=self.logger)
            return

        self.logger.info("Wiping benchmark candidate installation at [%s].", host.node.binary_path)

        for data_path in host.node.data_paths:
            self.path_manager.delete_path(host, data_path)

        self.path_manager.delete_path(host, host.node.binary_path)



import jinja2
from jinja2 import select_autoescape

from osbenchmark.exceptions import InvalidSyntax, SystemSetupError
from osbenchmark.utils import io


class TemplateRenderer:
    def render_template_file(self, root_path, variables, file_name):
        return self._handle_template_rendering_exceptions(self._render_template_file, root_path, variables, file_name)

    def _render_template_file(self, root_path, variables, file_name):
        env = jinja2.Environment(loader=jinja2.FileSystemLoader(root_path), autoescape=select_autoescape(['html', 'xml']))
        template = env.get_template(io.basename(file_name))
        # force a new line at the end. Jinja seems to remove it.
        return template.render(variables) + "\n"

    def render_template_string(self, template_string, variables):
        return self._handle_template_rendering_exceptions(self._render_template_string, template_string, variables)

    def _render_template_string(self, template_string, variables):
        env = jinja2.Environment(loader=jinja2.BaseLoader, autoescape=select_autoescape(['html', 'xml']))
        template = env.from_string(template_string)

        return template.render(variables)

    def _handle_template_rendering_exceptions(self, render_func, *args):
        try:
            return render_func(*args)
        except jinja2.exceptions.TemplateSyntaxError as e:
            raise InvalidSyntax("%s" % str(e))
        except BaseException as e:
            raise SystemSetupError("%s" % str(e))



import os

from osbenchmark.exceptions import SystemSetupError
from osbenchmark.utils.repo import BenchmarkRepository


class ConfigPathResolver:
    def __init__(self, cfg):
        self.cfg = cfg

    def resolve_config_path(self, config_type, config_version):
        config_root_path = self._get_config_root_path()
        root_path = os.path.join(config_root_path, config_type, f"v{config_version}")
        if not os.path.exists(root_path):
            raise SystemSetupError(f"Path {root_path} for {config_type} does not exist.")

        return root_path

    def _get_config_root_path(self):
        root_path = self.cfg.opts("builder", "provision_config.path", mandatory=False)
        if root_path:
            return root_path
        else:
            distribution_version = self.cfg.opts("builder", "distribution.version", mandatory=False)
            repo_name = self.cfg.opts("builder", "repository.name")
            repo_revision = self.cfg.opts("builder", "repository.revision")
            offline = self.cfg.opts("system", "offline.mode")
            default_directory = self.cfg.opts("provision_configs", "%s.dir" % repo_name, mandatory=False)
            root = self.cfg.opts("node", "root.dir")
            provision_config_repositories = self.cfg.opts("builder", "provision_config.repository.dir")
            provision_configs_dir = os.path.join(root, provision_config_repositories)

            current_provision_config_repo = BenchmarkRepository(
                default_directory, provision_configs_dir,
                repo_name, "provision_configs", offline)

            current_provision_config_repo.set_provision_configs_dir(repo_revision, distribution_version, self.cfg)
            return current_provision_config_repo.repo_dir



import logging
import os

from osbenchmark.builder.downloaders.downloader import Downloader
from osbenchmark.builder.utils.binary_keys import BinaryKeys


class OpenSearchSourceDownloader(Downloader):
    def __init__(self, provision_config_instance, executor, source_repository_provider, binary_builder, template_renderer,
                 artifact_variables_provider):
        super().__init__(executor)
        self.logger = logging.getLogger(__name__)
        self.provision_config_instance = provision_config_instance
        self.source_repository_provider = source_repository_provider
        self.binary_builder = binary_builder
        self.template_renderer = template_renderer
        self.artifact_variables_provider = artifact_variables_provider

    def download(self, host):
        opensearch_source_path = self._get_opensearch_source_path()
        self._fetch(host, opensearch_source_path)

        artifact_variables = self.artifact_variables_provider.get_artifact_variables(host)
        self._prepare(host, artifact_variables)

        return {BinaryKeys.OPENSEARCH: self._get_zip_path(opensearch_source_path, artifact_variables)}

    def _get_opensearch_source_path(self):
        node_root_dir = self.provision_config_instance.variables["source"]["root"]["dir"]
        opensearch_source_subdir = self.provision_config_instance.variables["source"]["opensearch"]["subdir"]
        return os.path.join(node_root_dir, opensearch_source_subdir)

    def _fetch(self, host, opensearch_source_path):
        plugin_remote_url = self.provision_config_instance.variables["source"]["remote"]["repo"]["url"]
        plugin_revision = self.provision_config_instance.variables["source"]["revision"]

        self.source_repository_provider.fetch_repository(host, plugin_remote_url, plugin_revision, opensearch_source_path)

    def _prepare(self, host, artifact_variables):
        clean_command_template = self.provision_config_instance.variables["source"]["clean"]["command"]
        build_command_template = self.provision_config_instance.variables["source"]["build"]["command"]

        if self.binary_builder:
            self.binary_builder.build(host, [
                self.template_renderer.render_template_string(clean_command_template, artifact_variables),
                self.template_renderer.render_template_string(build_command_template, artifact_variables)
            ])

    def _get_zip_path(self, opensearch_source_path, artifact_variables):
        artifact_path_pattern_template = self.provision_config_instance.variables["source"]["artifact_path_pattern"]
        artifact_path_pattern = self.template_renderer.render_template_string(artifact_path_pattern_template, artifact_variables)

        return os.path.join(opensearch_source_path, artifact_path_pattern)



# SPDX-License-Identifier: Apache-2.0
#
# The OpenSearch Contributors require contributions made to
# this file be licensed under the Apache-2.0 license or a
# compatible open source license.
# Modifications Copyright OpenSearch Contributors. See
# GitHub history for details.
# Licensed to Elasticsearch B.V. under one or more contributor
# license agreements. See the NOTICE file distributed with
# this work for additional information regarding copyright
# ownership. Elasticsearch B.V. licenses this file to you under
# the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#	http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import logging
import socket

import thespian.actors
import thespian.system.messages.status

from osbenchmark import exceptions, log
from osbenchmark.utils import console, net


class BenchmarkFailure:
    """
    Indicates a failure in the benchmark execution due to an exception
    """
    def __init__(self, message, cause=None):
        self.message = message
        self.cause = cause


class BenchmarkCancelled:
    """
    Indicates that the benchmark has been cancelled (by the user).
    """


def parametrized(decorator):
    """

    Helper meta-decorator that allows us to provide parameters to a decorator.

    :param decorator: The decorator that should accept parameters.
    """
    def inner(*args, **kwargs):
        def g(f):
            return decorator(f, *args, **kwargs)
        return g
    return inner


@parametrized
def no_retry(f, actor_name):
    """

    Decorator intended for Thespian message handlers with the signature ``receiveMsg_$MSG_NAME(self, msg, sender)``. Thespian will
    assume that a message handler that raises an exception can be retried. It will then retry once and give up afterwards just leaving
    a trace of that in the actor system's internal log file. However, this is usually *not* what we want in Benchmark. If handling of a
    message fails we instead want to notify a node higher up in the actor hierarchy.

    We achieve that by sending a ``BenchmarkFailure`` message to the original sender. Note that this might as well be the current
    actor (e.g. when handling a ``Wakeup`` message). In that case the actor itself is responsible for forwarding the benchmark failure
    to its parent actor.

    Example usage:

    @no_retry("special forces actor")
    def receiveMsg_DefuseBomb(self, msg, sender):
        # might raise an exception
        pass

    If this message handler raises an exception, the decorator will turn it into a ``BenchmarkFailure`` message with its ``message``
    property set to "Error in special forces actor" which is returned to the original sender.

    :param f: The message handler. Does not need to passed directly, this is handled by the decorator infrastructure.
    :param actor_name: A human readable name of the current actor that should be used in the exception message.
    """
    def guard(self, msg, sender):
        try:
            return f(self, msg, sender)
        except BaseException as e:
            msg = "Error in {}".format(actor_name)
            # log here as the full trace might get lost.
            logging.getLogger(__name__).exception(msg)
            # don't forward the exception as is because the main process might not have this class available on the load path
            # and will fail then while deserializing the cause.
            self.send(sender, BenchmarkFailure("{} ({})".format(msg, str(e))))
    return guard


class BenchmarkActor(thespian.actors.ActorTypeDispatcher):
    def __init__(self, *args, **kw):
        super().__init__(*args, **kw)
        self.children = []
        self.received_responses = []
        self.status = None
        log.post_configure_actor_logging()
        self.logger = logging.getLogger(__name__)
        console.set_assume_tty(assume_tty=False)

    # The method name is required by the actor framework
    # noinspection PyPep8Naming
    @staticmethod
    def actorSystemCapabilityCheck(capabilities, requirements):
        logger = logging.getLogger(__name__)
        for name, value in requirements.items():
            current = capabilities.get(name, None)
            if current != value:
                # A mismatch by is not a problem by itself as long as at least one actor system instance matches the requirements.
                logger.info("Checking capabilities [%s] against requirements [%s] failed.", capabilities, requirements)
                return False
        logger.info("Capabilities [%s] match requirements [%s].", capabilities, requirements)
        return True

    def transition_when_all_children_responded(self, sender, msg, expected_status, new_status, transition):
        """

        Waits until all children have sent a specific response message and then transitions this actor to a new status.

        :param sender: The child actor that has responded.
        :param msg: The response message.
        :param expected_status: The status in which this actor should be upon calling this method.
        :param new_status: The new status once all child actors have responded.
        :param transition: A parameter-less function to call immediately after changing the status.
        """
        if self.is_current_status_expected(expected_status):
            self.received_responses.append(msg)
            response_count = len(self.received_responses)
            expected_count = len(self.children)

            self.logger.debug("[%d] of [%d] child actors have responded for transition from [%s] to [%s].",
                              response_count, expected_count, self.status, new_status)
            if response_count == expected_count:
                self.logger.debug("All [%d] child actors have responded. Transitioning now from [%s] to [%s].",
                                  expected_count, self.status, new_status)
                # all nodes have responded, change status
                self.status = new_status
                self.received_responses = []
                transition()
            elif response_count > expected_count:
                raise exceptions.BenchmarkAssertionError(
                    "Received [%d] responses but only [%d] were expected to transition from [%s] to [%s]. The responses are: %s" %
                    (response_count, expected_count, self.status, new_status, self.received_responses))
        else:
            raise exceptions.BenchmarkAssertionError("Received [%s] from [%s] but we are in status [%s] instead of [%s]." %
                                                 (type(msg), sender, self.status, expected_status))

    def send_to_children_and_transition(self, sender, msg, expected_status, new_status):
        """

        Sends the provided message to all child actors and immediately transitions to the new status.

        :param sender: The actor from which we forward this message (in case it is message forwarding). Otherwise our own address.
        :param msg: The message to send.
        :param expected_status: The status in which this actor should be upon calling this method.
        :param new_status: The new status.
        """
        if self.is_current_status_expected(expected_status):
            self.logger.info("Transitioning from [%s] to [%s].", self.status, new_status)
            self.status = new_status
            for m in filter(None, self.children):
                self.send(m, msg)
        else:
            raise exceptions.BenchmarkAssertionError("Received [%s] from [%s] but we are in status [%s] instead of [%s]." %
                                                 (type(msg), sender, self.status, expected_status))

    def is_current_status_expected(self, expected_status):
        # if we don't expect anything, we're always in the right status
        if not expected_status:
            return True
        # do an explicit check for a list here because strings are also iterable and we have very tight control over this code anyway.
        elif isinstance(expected_status, list):
            return self.status in expected_status
        else:
            return self.status == expected_status


def actor_system_already_running(ip="127.0.0.1"):
    """
    Determines whether an actor system is already running by opening a socket connection.

    Note: It may be possible that another system is running on the same port.
    """
    s = socket.socket()
    try:
        s.connect((ip, 1900))
        s.close()
        return True
    except Exception:
        return False


__SYSTEM_BASE = "multiprocTCPBase"


def use_offline_actor_system():
    global __SYSTEM_BASE
    __SYSTEM_BASE = "multiprocQueueBase"


def bootstrap_actor_system(try_join=False, prefer_local_only=False, local_ip=None, coordinator_ip=None):
    logger = logging.getLogger(__name__)
    system_base = __SYSTEM_BASE
    try:
        if try_join:
            if actor_system_already_running():
                logger.info("Joining already running actor system with system base [%s].", system_base)
                return thespian.actors.ActorSystem(system_base)
            else:
                logger.info("Creating new actor system with system base [%s] on coordinator node.", system_base)
                # if we try to join we can only run on the coordinator...
                return thespian.actors.ActorSystem(system_base, logDefs=log.load_configuration(), capabilities={"coordinator": True})
        elif prefer_local_only:
            coordinator = True
            if system_base != "multiprocQueueBase":
                coordinator_ip = "127.0.0.1"
                local_ip = "127.0.0.1"
            else:
                coordinator_ip = None
                local_ip = None
        else:
            if system_base not in ("multiprocTCPBase", "multiprocUDPBase"):
                raise exceptions.SystemSetupError("Benchmark requires a network-capable system base but got [%s]." % system_base)
            if not coordinator_ip:
                raise exceptions.SystemSetupError("coordinator IP is required")
            if not local_ip:
                raise exceptions.SystemSetupError("local IP is required")
            # always resolve the public IP here, even if a DNS name is given. Otherwise Thespian will be unhappy
            local_ip = net.resolve(local_ip)
            coordinator_ip = net.resolve(coordinator_ip)

            coordinator = local_ip == coordinator_ip

        capabilities = {"coordinator": coordinator}
        if local_ip:
            # just needed to determine whether to run benchmarks locally
            capabilities["ip"] = local_ip
        if coordinator_ip:
            # Make the coordinator node the convention leader
            capabilities["Convention Address.IPv4"] = "%s:1900" % coordinator_ip
        logger.info("Starting actor system with system base [%s] and capabilities [%s].", system_base, capabilities)
        return thespian.actors.ActorSystem(system_base,
                                           logDefs=log.load_configuration(),
                                           capabilities=capabilities)
    except thespian.actors.ActorSystemException:
        logger.exception("Could not initialize internal actor system.")
        raise
